r"""
.. _vqe:

Variational quantum eigensolver
===============================

This example demonstrates the principle of a variational quantum
eigensolver (VQE), originally proposed in `Peruzzo et al.
(2014) <https://www.nature.com/articles/ncomms5213>`__. To showcase the
hybrid computational capabilities of PennyLane, we first train a quantum
circuit to minimize the squared energy expectation for a Hamiltonian
:math:`H`,

.. math::

    \langle \psi_v | H | \psi_v \rangle^2  =( 0.1 \langle \psi_{v} | X_2 |
    \psi_v \rangle + 0.5 \langle \psi_v | Y_2 | \psi_v \rangle )^2.

Here, :math:`|\psi_v\rangle` is the state
obtained after applying a quantum circuit to an initial state
:math:`|0\rangle`. The quantum circuit depends on trainable variables
:math:`v = \{v_1, v_2\}`, and :math:`X_2`, :math:`Y_2` denote the
Pauli-X and Pauli-Y operator acting on the second qubit (*Note: We apply
the square to make the optimization landscapes more interesting, but in
common applications the cost is directly the energy expectation value*).

After doing this, we will then turn things around and use a fixed
quantum circuit to prepare a state :math:`|\psi\rangle`, but train the coefficients of
the Hamiltonian to minimize

.. math::

    \langle \psi | H | \psi \rangle^2  = (v_1 \langle \psi | X_2 | \psi
    \rangle + v_2 \langle \psi | Y_2 | \psi \rangle )^2 .
"""

##############################################################################
# 1. Optimizing the quantum circuit
# ---------------------------------
#
# Imports
# ~~~~~~~
#
# We begin by importing PennyLane, the PennyLane-wrapped version of NumPy,
# and the GradientDescentOptimizer.

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import GradientDescentOptimizer

##############################################################################
# We use the default qubit simulator as a device.

dev = qml.device("default.qubit", wires=2)

##############################################################################
# Quantum nodes
# ~~~~~~~~~~~~~
#
# The quantum circuit of the variational eigensolver is an ansatz that
# defines a manifold of possible quantum states. We use a Hadamard, two
# rotations and a CNOT gate to construct our circuit.


def ansatz(var):
    qml.Rot(0.3, 1.8, 5.4, wires=1)
    qml.RX(var[0], wires=0)
    qml.RY(var[1], wires=1)
    qml.CNOT(wires=[0, 1])


##############################################################################
# A variational eigensolver requires us to evaluate expectations of
# different Pauli operators. In this example, the Hamiltonian is expressed
# by only two single-qubit Pauli operators, namely the X and Y operator
# applied to the first qubit.
#
# Since these operators will be measured on the same wire, we will need to
# create two quantum nodes (one for each operator whose expectation value
# we measure), but we can reuse the same device.
#
# .. note::
#
#     If the Pauli observables were evaluated on different wires, we
#     could use one quantum node and return a tuple of expectations in only
#     one quantum node:
#     ``return qml.expectation.PauliX(0), qml.expectation.PauliY(1)``


@qml.qnode(dev)
def circuit_X(var):
    ansatz(var)
    return qml.expval(qml.PauliX(1))


@qml.qnode(dev)
def circuit_Y(var):
    ansatz(var)
    return qml.expval(qml.PauliY(1))


##############################################################################
# Objective
# ~~~~~~~~~

# The cost function to be optimized in VQE is simply a linear combination
# of the expectations, which defines the expectation of the Hamiltonian we
# are interested in. In our case, we square this cost function to provide
# a more interesting landscape with the same minima.


def cost(var):
    expX = circuit_X(var)
    expY = circuit_Y(var)
    return (0.1 * expX + 0.5 * expY) ** 2


##############################################################################
# This cost defines the following landscape:
#
# *Note: To run the following cell you need the matplotlib library.*

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from matplotlib.ticker import MaxNLocator

fig = plt.figure(figsize=(6, 4))
ax = fig.gca(projection="3d")

X = np.linspace(-3.0, 3.0, 20)
Y = np.linspace(-3.0, 3.0, 20)
xx, yy = np.meshgrid(X, Y)
Z = np.array([[cost([x, y]) for x in X] for y in Y]).reshape(len(Y), len(X))
surf = ax.plot_surface(xx, yy, Z, cmap=cm.coolwarm, antialiased=False)

ax.set_xlabel("v1")
ax.set_ylabel("v2")
ax.zaxis.set_major_locator(MaxNLocator(nbins=5, prune="lower"))

plt.show()

##############################################################################
# Optimization
# ~~~~~~~~~~~~
#
# We create a GradientDescentOptimizer and use it to optimize the cost
# function.

opt = GradientDescentOptimizer(0.5)

var = [0.3, 2.5]
var_gd = [var]
for it in range(20):
    var = opt.step(cost, var)
    var_gd.append(var)

    print(
        "Cost after step {:5d}: {: .7f} | Variables: [{: .5f},{: .5f}]".format(
            it + 1, cost(var), var[0], var[1]
        )
    )

##############################################################################
# We can plot the path that the variables took during gradient descent. To
# make the plot more clear, we will shorten the range for :math:`v_2`.

fig = plt.figure(figsize=(6, 4))
ax = fig.gca(projection="3d")

X = np.linspace(-3, np.pi / 2, 20)
Y = np.linspace(-3, 3, 20)
xx, yy = np.meshgrid(X, Y)
Z = np.array([[cost([x, y]) for x in X] for y in Y]).reshape(len(Y), len(X))

surf = ax.plot_surface(xx, yy, Z, cmap=cm.coolwarm, antialiased=False)

path_z = [cost(var) + 1e-8 for var in var_gd]
path_x = [v[0] for v in var_gd]
path_y = [v[1] for v in var_gd]
ax.plot(path_x, path_y, path_z, c="green", marker=".", label="graddesc", zorder=10)

ax.set_xlabel("v1")
ax.set_ylabel("v2")
ax.zaxis.set_major_locator(MaxNLocator(nbins=5, prune="lower"))

plt.legend()
plt.show()


##############################################################################
# 2. Optimizing the Hamiltonian coefficients
# ------------------------------------------
#
# Instead of optimizing the circuit parameters, we can also use a fixed
# circuit,


def ansatz():
    qml.Rot(0.3, 1.8, 5.4, wires=1)
    qml.RX(-0.5, wires=0)
    qml.RY(0.5, wires=1)
    qml.CNOT(wires=[0, 1])


@qml.qnode(dev)
def circuit_X():
    ansatz()
    return qml.expval(qml.PauliX(1))


@qml.qnode(dev)
def circuit_Y():
    ansatz()
    return qml.expval(qml.PauliY(1))


##############################################################################
# and make the classical coefficients that appear in the Hamiltonian the
# trainable variables.


def cost(var):
    expX = circuit_X()
    expY = circuit_Y()
    return (var[0] * expX + var[1] * expY) ** 2


opt = GradientDescentOptimizer(0.5)

var = [0.3, 2.5]
var_gd = [var]
for it in range(20):
    var = opt.step(cost, var)
    var_gd.append(var)

    print(
        "Cost after step {:5d}: {: .7f} | Variables: [{: .5f},{: .5f}]".format(
            it + 1, cost(var), var[0], var[1]
        )
    )

##############################################################################
# The landscape has a quadratic shape.

fig = plt.figure(figsize=(6, 4))
ax = fig.gca(projection="3d")

X = np.linspace(-3, np.pi / 2, 20)
Y = np.linspace(-3, 3, 20)
xx, yy = np.meshgrid(X, Y)
Z = np.array([[cost([x, y]) for x in X] for y in Y]).reshape(len(Y), len(X))
surf = ax.plot_surface(xx, yy, Z, cmap=cm.coolwarm, antialiased=False)

path_z = [cost(var) + 1e-8 for var in var_gd]
path_x = [v[0] for v in var_gd]
path_y = [v[1] for v in var_gd]
ax.plot(path_x, path_y, path_z, c="pink", marker=".", label="graddesc", zorder=10)

ax.set_xlabel("v1")
ax.set_ylabel("v2")
ax.zaxis.set_major_locator(MaxNLocator(nbins=5, prune="lower"))

plt.legend()
plt.show()


##############################################################################
# 3. Optimizing classical and quantum parameters
# ----------------------------------------------

# Finally, we can optimize *classical* and *quantum* weights together by
# combining the two approaches from above.


def ansatz(var):

    qml.Rot(0.3, 1.8, 5.4, wires=1)
    qml.RX(var[0], wires=0)
    qml.RY(var[1], wires=1)
    qml.CNOT(wires=[0, 1])


@qml.qnode(dev)
def circuit_X(var):
    ansatz(var)
    return qml.expval(qml.PauliX(1))


@qml.qnode(dev)
def circuit_Y(var):
    ansatz(var)
    return qml.expval(qml.PauliY(1))


def cost(var):

    expX = circuit_X(var)
    expY = circuit_Y(var)

    return (var[2] * expX + var[3] * expY) ** 2


opt = GradientDescentOptimizer(0.5)
var = [0.3, 2.5, 0.3, 2.5]

for it in range(10):
    var = opt.step(cost, var)
    print("Cost after step {:5d}: {: 0.7f}".format(it + 1, cost(var)))



r"""
.. _qubit_rotation:

Basic tutorial: qubit rotation
==============================

To see how PennyLane allows the easy construction and optimization of quantum functions, let's
consider the simple case of **qubit rotation** the PennyLane version of the 'Hello, world!'
example.

The task at hand is to optimize two rotation gates in order to flip a single
qubit from state :math:`\left|0\right\rangle` to state :math:`\left|1\right\rangle`.


The quantum circuit
-------------------

In the qubit rotation example, we wish to implement the following quantum circuit:

.. figure:: ../../examples/figures/rotation_circuit.png
    :align: center
    :width: 40%
    :target: javascript:void(0);

Breaking this down step-by-step, we first start with a qubit in the ground state
:math:`|0\rangle = \begin{bmatrix}1 & 0 \end{bmatrix}^T`,
and rotate it around the x-axis by applying the gate

.. math::
    R_x(\phi_1) = e^{-i \phi_1 \sigma_x /2} =
    \begin{bmatrix} \cos \frac{\phi_1}{2} &  -i \sin \frac{\phi_1}{2} \\
                   -i \sin \frac{\phi_1}{2} &  \cos \frac{\phi_1}{2}
    \end{bmatrix},

and then around the y-axis via the gate

.. math::
    R_y(\phi_2) = e^{-i \phi_2 \sigma_y/2} =
   \begin{bmatrix} \cos \frac{\phi_2}{2} &  - \sin \frac{\phi_2}{2} \\
                   \sin \frac{\phi_2}{2} &  \cos \frac{\phi_2}{2}
   \end{bmatrix}.

After these operations the qubit is now in the state

.. math::  | \psi \rangle = R_y(\phi_2) R_x(\phi_1) | 0 \rangle.

Finally, we measure the expectation value :math:`\langle \psi \mid \sigma_z \mid \psi \rangle`
of the Pauli-Z operator

.. math::
   \sigma_z =
   \begin{bmatrix} 1 &  0 \\
                   0 & -1
   \end{bmatrix}.

Using the above to calculate the exact expectation value, we find that

.. math::
    \langle \psi \mid \sigma_z \mid \psi \rangle
    = \langle 0 \mid R_x(\phi_1)^\dagger R_y(\phi_2)^\dagger \sigma_z  R_y(\phi_2) R_x(\phi_1) \mid 0 \rangle
    = \cos(\phi_1)\cos(\phi_2).

Depending on the circuit parameters :math:`\phi_1` and :math:`\phi_2`, the
output expectation lies between :math:`1` (if :math:`\left|\psi\right\rangle = \left|0\right\rangle`)
and :math:`-1` (if :math:`\left|\psi\right\rangle = \left|1\right\rangle`).

Let's see how we can easily implement and optimize this circuit using PennyLane.
"""

##############################################################################
# Importing PennyLane and NumPy
# -----------------------------
#
# The first thing we need to do is import PennyLane, as well as the wrapped version
# of NumPy provided by PennyLane.

import pennylane as qml
from pennylane import numpy as np


##############################################################################
# .. important::
#
#     When constructing a hybrid quantum/classical computational model with PennyLane,
#     it is important to **always import NumPy from PennyLane**, not the standard NumPy!
#
#     By importing the wrapped version of NumPy provided by PennyLane, you can combine
#     the power of NumPy with PennyLane:
#
#     * continue to use the classical NumPy functions and arrays you know and love
#     * combine quantum functions (evaluated on quantum hardware/simulators) and
#       classical functions (provided by NumPy)
#     * allow PennyLane to automatically calculate gradients of both classical and
#       quantum functions


##############################################################################
# Creating a device
# -----------------
#
# Before we can construct our quantum node, we need to initialize a **device**.
#
# .. admonition:: Definition
#     :class: defn
#
#     Any computational object that can apply quantum operations, and return an measurement value
#     is called a quantum **device**.
#
#     In PennyLane, a device could be a hardware device (such as the IBM QX4, via the
#     PennyLane-PQ plugin), or a software simulator (such as Strawberry Fields, via the
#     PennyLane-SF plugin).
#
# .. tip::
#
#    *Devices are loaded in PennyLane via the function* :func:`pennylane.device`
#
#
# PennyLane supports devices using both the qubit model of quantum computation and devices
# using the CV model of quantum computation. In fact, even a hybrid computation containing
# both qubit and CV quantum nodes is possible; see the
# :ref:`hybrid computation example <hybrid_computation_example>` for more details.
#
# For this tutorial, we are using the qubit model, so let's initialize the ``'default.qubit'`` device
# provided by PennyLane; a simple pure-state qubit simulator.

dev1 = qml.device("default.qubit", wires=1)

##############################################################################
# For all devices, :func:`~.pennylane.device` accepts the following arguments:
#
# * ``name``: the name of the device to be loaded
# * ``wires``: the number of subsystems to initialize the device with
#
# Here, as we only require a single qubit for this example, we set ``wires=1``.

##############################################################################
# Constructing the QNode
# ----------------------
#
# Now that we have initialized our device, we can begin to construct a
# **quantum node** (or QNode).
#
#
# .. admonition:: Definition
#     :class: defn
#
#     QNodes are an abstract encapsulation of a quantum function, described by a
#     quantum circuit. QNodes are bound to a particular quantum device, which is
#     used to evaluate expectation and variance values of this circuit.
#
# .. tip::
#
#    *QNodes can be constructed via the* :class:`pennylane.QNode <pennylane.qnode.QNode>`
#    *class, or by using the provided* :mod:`qnode decorator <pennylane.decorator>`.
#
#
# First, we need to define the quantum function that will be evaluated in the QNode:


def circuit(params):
    qml.RX(params[0], wires=0)
    qml.RY(params[1], wires=0)
    return qml.expval(qml.PauliZ(0))


##############################################################################
# This is a simple circuit, matching the one described above.
# Notice that the function ``circuit()`` is constructed as if it were any
# other Python function; it accepts a positional argument ``params``, which may
# be a list, tuple, or array, and uses the individual elements for gate parameters.
#
# However, quantum functions are a **restricted subset** of Python functions.
# For a Python function to also be a valid quantum function, there are some
# important restrictions:
#
# * **Quantum functions must only contain quantum operations, one operation per
#   line, in the order in which they are to be applied.**
#
#   In addition, we must always specify the subsystem the operation applies to,
#   by passing the ``wires`` argument; this may be a list or an integer, depending
#   on how many wires the operation acts on.
#
#   For a full list of quantum operations, see :mod:`supported operations <pennylane.ops>`.
#
# * **Quantum functions must return either a single or a tuple of measured observables**.
#
#   As a result, the quantum function always returns a classical quantity, allowing
#   the QNode to interface with other classical functions (and also other QNodes).
#
#   See the full list of :mod:`quantum observables <pennylane.ops>`, as well
#   supported :mod:`measurement return types <pennylane.measure>`.
#
# * **Quantum functions must not contain any classical processing of circuit parameters.**
#
# .. note::
#
#     Certain devices may only support a subset of the available PennyLane
#     operations/observables, or may even provide additional operations/observables.
#     Please consult the documentation for the plugin/device for more details.
#
# Once we have written the quantum function, we convert it into a :class:`~.QNode` running
# on device ``dev1`` by applying the :mod:`qnode decorator <pennylane.decorator>`
# **directly above** the function definition:


@qml.qnode(dev1)
def circuit(params):
    qml.RX(params[0], wires=0)
    qml.RY(params[1], wires=0)
    return qml.expval(qml.PauliZ(0))


##############################################################################
# Thus, our ``circuit()`` quantum function is now a :class:`~.QNode`, which will run on
# device ``dev1`` every time it is evaluated.
#
# To evaluate, we simply call the function with some appropriate numerical inputs:

print(circuit([0.54, 0.12]))

##############################################################################
# Calculating quantum gradients
# -----------------------------
#
# The gradient of the function ``circuit``, encapsulated within the ``QNode``,
# can be evaluated by utilizing the same quantum
# device (``dev1``) that we used to evaluate the function itself.
#
# PennyLane incorporates both analytic differentiation, as well as numerical
# methods (such as the method of finite differences). Both of these are done
# automatically.
#
# We can differentiate by using the built-in :func:`~.pennylane.grad` function.
# This returns another function, representing the gradient (i.e., the vector of
# partial derivatives) of ``circuit``. The gradient can be evaluated in the same
# way as the original function:

dcircuit = qml.grad(circuit, argnum=0)

##############################################################################
# The function :func:`~.pennylane.grad` itself **returns a function**, representing
# the derivative of the QNode with respect to the argument specified in ``argnum``.
# In this case, the function ``circuit`` takes one argument (``params``), so we
# specify ``argnum=0``. Because the argument has two elements, the returned gradient
# is two-dimensional. We can then evaluate this gradient function at any point in the parameter space.

print(dcircuit([0.54, 0.12]))

################################################################################
# **A note on arguments**
#
# Quantum circuit functions, being a restricted subset of Python functions,
# can also make use of multiple positional arguments and keyword arguments.
# For example, we could have defined the above quantum circuit function using
# two positional arguments, instead of one array argument:


@qml.qnode(dev1)
def circuit2(phi1, phi2):
    qml.RX(phi1, wires=0)
    qml.RY(phi2, wires=0)
    return qml.expval(qml.PauliZ(0))


################################################################################
# When we calculate the gradient for such a function, the usage of ``argnum``
# will be slightly different. In this case, ``argnum=0`` will return the gradient
# with respect to only the first parameter (``phi1``), and ``argnum=1`` will give
# the gradient for ``phi2``. To get the gradient with respect to both parameters,
# we can use ``argnum=[0,1]``:

dcircuit = qml.grad(circuit2, argnum=[0, 1])
print(dcircuit(0.54, 0.12))

################################################################################
# Keyword arguments may also be used in your custom quantum function. PennyLane
# does **not** differentiate QNodes with respect to keyword arguments,
# so they are useful for passing external data to your QNode.


################################################################################
# Optimization
# ------------
#
# .. admonition:: Definition
#     :class: defn
#
#     If using the default NumPy/Autograd interface, PennyLane provides a collection
#     of optimizers based on gradient descent. These optimizers accept a cost function
#     and initial parameters, and utilize PennyLane's automatic differentiation
#     to perform gradient descent.
#
# .. tip::
#
#    *See* :mod:`pennylane.optimize` *for details and documentation of available optimizers*
#
# Next, let's make use of PennyLane's built-in optimizers to optimize the two circuit
# parameters :math:`\phi_1` and :math:`\phi_2` such that the qubit, originally in state
# :math:`\left|0\right\rangle`, is rotated to be in state :math:`\left|1\right\rangle`. This is equivalent to measuring a
# Pauli-Z expectation value of :math:`-1`, since the state :math:`\left|1\right\rangle` is an eigenvector
# of the Pauli-Z matrix with eigenvalue :math:`\lambda=-1`.
#
# In other words, the optimization procedure will find the weights
# :math:`\phi_1` and :math:`\phi_2` that result in the following rotation on the Bloch sphere:
#
# .. figure:: ../../examples/figures/bloch.png
#     :align: center
#     :width: 70%
#     :target: javascript:void(0);
#
# To do so, we need to define a **cost** function. By *minimizing* the cost function, the
# optimizer will determine the values of the circuit parameters that produce the desired outcome.
#
# In this case, our desired outcome is a Pauli-Z expectation value of :math:`-1`. Since we
# know that the Pauli-Z expectation is bound between :math:`[-1, 1]`, we can define our
# cost directly as the output of the QNode:


def cost(var):
    return circuit(var)


################################################################################
# To begin our optimization, let's choose small initial values of :math:`\phi_1` and :math:`\phi_2`:

init_params = np.array([0.011, 0.012])
print(cost(init_params))

################################################################################
# We can see that, for these initial parameter values, the cost function is close to :math:`1`.
#
# Finally, we use an optimizer to update the circuit parameters for 100 steps. We can use the built-in
# :class:`pennylane.optimize.GradientDescentOptimizer` class:

# initialise the optimizer
opt = qml.GradientDescentOptimizer(stepsize=0.4)

# set the number of steps
steps = 100
# set the initial parameter values
params = init_params

for i in range(steps):
    # update the circuit parameters
    params = opt.step(cost, params)

    if (i + 1) % 5 == 0:
        print("Cost after step {:5d}: {: .7f}".format(i + 1, cost(params)))

print("Optimized rotation angles: {}".format(params))

################################################################################
# We can see that the optimization converges after approximately 40 steps.
#
# Substituting this into the theoretical result :math:`\langle \psi \mid \sigma_z \mid \psi \rangle = \cos\phi_1\cos\phi_2`,
# we can verify that this is indeed one possible value of the circuit parameters that
# produces :math:`\langle \psi \mid \sigma_z \mid \psi \rangle=-1`, resulting in the qubit being rotated
# to the state :math:`\left|1\right\rangle`.
#
# .. note::
#
#     Some optimizers, such as :class:`~.pennylane.optimize.AdagradOptimizer`, have
#     internal hyperparameters that are stored in the optimizer instance. These can
#     be reset using the :meth:`reset` method.
#
# Continue on to the next tutorial, :ref:`gaussian_transformation`, to see a similar example using
# continuous-variable (CV) quantum nodes.



"""
.. _quantum_GAN:

Quantum Generative Adversarial Network
======================================

This demo constructs a Quantum Generative Adversarial Network (QGAN)
(`Lloyd and Weedbrook
(2018) <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.121.040502>`__,
`Dallaire-Demers and Killoran
(2018) <https://journals.aps.org/pra/abstract/10.1103/PhysRevA.98.012324>`__)
using two subcircuits, a *generator* and a *discriminator*. The
generator attempts to generate synthetic quantum data to match a pattern
of “real” data, while the discriminator, tries to discern real data from
fake data; see image below. The gradient of the discriminator’s output provides a
training signal for the generator to improve its fake generated data.

|

.. figure:: ../../examples/figures/qgan.png
    :align: center
    :width: 75%
    :target: javascript:void(0)

|


"""

##############################################################################
# Imports
# ~~~~~~~

# As usual, we import PennyLane, the PennyLane-provided version of NumPy,
# and an optimizer.

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import GradientDescentOptimizer

##############################################################################
# We also declare a 3-qubit device.


dev = qml.device("default.qubit", wires=3)

##############################################################################
# Classical and quantum nodes
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In classical GANs, the starting point is to draw samples either from
# some “real data” distribution, or from the generator, and feed them to
# the discriminator. In this QGAN example, we will use a quantum circuit
# to generate the real data.
#
# For this simple example, our real data will be a qubit that has been
# rotated (from the starting state :math:`\left|0\right\rangle`) to some
# arbitrary, but fixed, state.


def real(phi, theta, omega):
    qml.Rot(phi, theta, omega, wires=0)


##############################################################################
# For the generator and discriminator, we will choose the same basic
# circuit structure, but acting on different wires.
#
# Both the real data circuit and the generator will output on wire 0,
# which will be connected as an input to the discriminator. Wire 1 is
# provided as a workspace for the generator, while the discriminator’s
# output will be on wire 2.


def generator(w):
    qml.RX(w[0], wires=0)
    qml.RX(w[1], wires=1)
    qml.RY(w[2], wires=0)
    qml.RY(w[3], wires=1)
    qml.RZ(w[4], wires=0)
    qml.RZ(w[5], wires=1)
    qml.CNOT(wires=[0, 1])
    qml.RX(w[6], wires=0)
    qml.RY(w[7], wires=0)
    qml.RZ(w[8], wires=0)


def discriminator(w):
    qml.RX(w[0], wires=0)
    qml.RX(w[1], wires=2)
    qml.RY(w[2], wires=0)
    qml.RY(w[3], wires=2)
    qml.RZ(w[4], wires=0)
    qml.RZ(w[5], wires=2)
    qml.CNOT(wires=[1, 2])
    qml.RX(w[6], wires=2)
    qml.RY(w[7], wires=2)
    qml.RZ(w[8], wires=2)


##############################################################################
# We create two QNodes. One where the real data source is wired up to the
# discriminator, and one where the generator is connected to the
# discriminator.


@qml.qnode(dev)
def real_disc_circuit(phi, theta, omega, disc_weights):
    real(phi, theta, omega)
    discriminator(disc_weights)
    return qml.expval(qml.PauliZ(2))


@qml.qnode(dev)
def gen_disc_circuit(gen_weights, disc_weights):
    generator(gen_weights)
    discriminator(disc_weights)
    return qml.expval(qml.PauliZ(2))


##############################################################################
# Cost
# ~~~~
#
# There are two ingredients to the cost here. The first is the probability
# that the discriminator correctly classifies real data as real. The
# second ingredient is the probability that the discriminator classifies
# fake data (i.e., a state prepared by the generator) as real.
#
# The discriminator’s objective is to maximize the probability of
# correctly classifying real data, while minimizing the probability of
# mistakenly classifying fake data.
#
# The generator’s objective is to maximize the probability that the
# discriminator accepts fake data as real.


def prob_real_true(disc_weights):
    true_disc_output = real_disc_circuit(phi, theta, omega, disc_weights)
    # convert to probability
    prob_real_true = (true_disc_output + 1) / 2
    return prob_real_true


def prob_fake_true(gen_weights, disc_weights):
    fake_disc_output = gen_disc_circuit(gen_weights, disc_weights)
    # convert to probability
    prob_fake_true = (fake_disc_output + 1) / 2
    return prob_fake_true  # generator wants to minimize this prob


def disc_cost(disc_weights):
    cost = prob_fake_true(gen_weights, disc_weights) - prob_real_true(disc_weights)
    return cost


def gen_cost(gen_weights):
    return -prob_fake_true(gen_weights, disc_weights)


##############################################################################
# Optimization
# ~~~~~~~~~~~~
#
# We initialize the fixed angles of the “real data” circuit, as well as
# the initial parameters for both generator and discriminator. These are
# chosen so that the generator initially prepares a state on wire 0 that
# is very close to the :math:`\left| 1 \right\rangle` state.

phi = np.pi / 6
theta = np.pi / 2
omega = np.pi / 7
np.random.seed(0)
eps = 1e-2
gen_weights = np.array([np.pi] + [0] * 8) + np.random.normal(scale=eps, size=[9])
disc_weights = np.random.normal(size=[9])

##############################################################################
# We begin by creating the optimizer:

opt = GradientDescentOptimizer(0.1)

##############################################################################
# In the first stage of training, we optimize the discriminator while
# keeping the generator parameters fixed.

for it in range(50):
    disc_weights = opt.step(disc_cost, disc_weights)
    cost = disc_cost(disc_weights)
    if it % 5 == 0:
        print("Step {}: cost = {}".format(it + 1, cost))

##############################################################################
# At the discriminator’s optimum, the probability for the discriminator to
# correctly classify the real data should be close to one.

print(prob_real_true(disc_weights))


##############################################################################
# For comparison, we check how the discriminator classifies the
# generator’s (still unoptimized) fake data:

print(prob_fake_true(gen_weights, disc_weights))


##############################################################################
# In the adverserial game we have to now train the generator to better
# fool the discriminator (we can continue training the models in an
# alternating fashion until we reach the optimum point of the two-player
# adversarial game).

for it in range(200):
    gen_weights = opt.step(gen_cost, gen_weights)
    cost = -gen_cost(gen_weights)
    if it % 5 == 0:
        print("Step {}: cost = {}".format(it, cost))


##############################################################################
# At the optimum of the generator, the probability for the discriminator
# to be fooled should be close to 1.

print(prob_fake_true(gen_weights, disc_weights))


##############################################################################
# At the joint optimum the overall cost will be close to zero.

print(disc_cost(disc_weights))


##############################################################################
# The generator has successfully learned how to simulate the real data
# enough to fool the discriminator.



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=protected-access
r"""
This module contains the abstract base classes for defining PennyLane
operations and observables.

Description
-----------

Qubit Operations
----------------
The :class:`Operator` class serves as a base class for operators,
and is inherited by both the :class:`Observable` class and the
:class:`Operation` class. These classes are subclassed to implement quantum operations
and measure observables in PennyLane.

* Each :class:`~.Operator` subclass represents a general type of
  map between physical states. Each instance of these subclasses
  represents either

  - an application of the operator or
  - an instruction to measure and return the respective result.

  Operators act on a sequence of wires (subsystems) using given parameter values.

* Each :class:`~.Operation` subclass represents a type of quantum operation,
  for example a unitary quantum gate. Each instance of these subclasses
  represents an application of the operation with given parameter values to
  a given sequence of wires (subsystems).

* Each  :class:`~.Observable` subclass represents a type of physical observable.
  Each instance of these subclasses represents an instruction to measure and
  return the respective result for the given parameter values on a
  sequence of wires (subsystems).

Differentiation
^^^^^^^^^^^^^^^

In general, an :class:`Operation` is differentiable (at least using the finite-difference
method) with respect to a parameter iff

* the domain of that parameter is continuous.

For an :class:`Operation` to be differentiable with respect to a parameter using the
analytic method of differentiation, it must satisfy an additional constraint:

* the parameter domain must be real.

.. note::

    These conditions are *not* sufficient for analytic differentiation. For example,
    CV gates must also define a matrix representing their Heisenberg linear
    transformation on the quadrature operators.

For gates that *are* supported via the analytic method, the gradient recipe
(with multiplier :math:`c_k`, parameter shift :math:`s_k` for parameter :math:`\phi_k`)
works as follows:

.. math:: \frac{\partial}{\partial\phi_k}O = c_k\left[O(\phi_k+s_k)-O(\phi_k-s_k)\right].

CV Operation base classes
~~~~~~~~~~~~~~~~~~~~~~~~~

Due to additional requirements, continuous-variable (CV) operations must subclass the
:class:`~.CVOperation` or :class:`~.CVObservable` classes instead of :class:`~.Operation`
and :class:`~.Observable`.

Differentiation
^^^^^^^^^^^^^^^

To enable gradient computation using the analytic method for Gaussian CV operations, in addition, you need to
provide the static class method :meth:`~.CV._heisenberg_rep` that returns the Heisenberg representation of
the operation given its list of parameters, namely:

* For Gaussian CV Operations this method should return the matrix of the linear transformation carried out by the
  operation on the vector of quadrature operators :math:`\mathbf{r}` for the given parameter
  values.

* For Gaussian CV Observables this method should return a real vector (first-order observables)
  or symmetric matrix (second-order observables) of coefficients of the quadrature
  operators :math:`\x` and :math:`\p`.

PennyLane uses the convention :math:`\mathbf{r} = (\I, \x, \p)` for single-mode operations and observables
and :math:`\mathbf{r} = (\I, \x_0, \p_0, \x_1, \p_1, \ldots)` for multi-mode operations and observables.

.. note::
    Non-Gaussian CV operations and observables are currently only supported via
    the finite-difference method of gradient computation.
"""
import abc
from enum import Enum, IntEnum
import numbers
from collections.abc import Sequence

import numpy as np

from .qnode import QNode
from .utils import _flatten, _unflatten
from .variable import Variable


#=============================================================================
# Wire types
#=============================================================================

class Wires(IntEnum):
    """Integer enumeration class
    to represent the number of wires
    an operation acts on"""
    Any = -1
    All = 0


All = Wires.All
"""IntEnum: An enumeration which represents all wires in the
subsystem. It is equivalent to an integer with value 0."""

Any = Wires.Any
"""IntEnum: An enumeration which represents any wires in the
subsystem. It is equivalent to an integer with value -1."""


#=============================================================================
# ObservableReturnTypes types
#=============================================================================

class ObservableReturnTypes(Enum):
    """Enumeration class to
    represent the type of
    return types of an observable."""
    Sample = 1
    Variance = 2
    Expectation = 3


Sample = ObservableReturnTypes.Sample
"""Enum: An enumeration which represents sampling an observable."""

Variance = ObservableReturnTypes.Variance
"""Enum: An enumeration which represents returning the variance of
an observable on specified wires."""

Expectation = ObservableReturnTypes.Expectation
"""Enum: An enumeration which represents returning the expectation
value of an observable on specified wires."""


#=============================================================================
# Class property
#=============================================================================


class ClassPropertyDescriptor: # pragma: no cover
    """Allows a class property to be defined"""
    # pylint: disable=too-few-public-methods
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        """Set the function as a class method, and store as an attribute."""
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    """The class property decorator"""
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


#=============================================================================
# Base Operator class
#=============================================================================

class Operator(abc.ABC):
    r"""Base class for quantum operators supported by a device.

    The following class attributes must be defined for all Operators:

    * :attr:`~.Operator.num_params`
    * :attr:`~.Operator.num_wires`
    * :attr:`~.Operator.par_domain`

    Args:
        args (tuple[float, int, array, Variable]): operator parameters

    Keyword Args:
        wires (Sequence[int]): Subsystems it acts on. If not given, args[-1]
            is interpreted as wires.
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into a :class:`QNode` circuit queue.
            The circuit queue is determined by the presence of an
            applicable `QNode._current_context`. If no context is
            available, this argument is ignored.
    """

    @property
    @abc.abstractmethod
    def num_params(self):
        """Number of parameters the operator takes."""

    @property
    @abc.abstractmethod
    def num_wires(self):
        """Number of wires the operator acts on.

        The value 0 allows the operator to act on any number of wires.
        """

    @property
    @abc.abstractmethod
    def par_domain(self):
        """Domain of the gate parameters.

        * ``'N'``: natural numbers (including zero).
        * ``'R'``: floats.
        * ``'A'``: arrays of real or complex values.
        * ``None``: if there are no parameters.
        """

    def __init__(self, *args, wires=None, do_queue=True):
        # pylint: disable=too-many-branches
        self.name = self.__class__.__name__   #: str: name of the operation
        self.queue_idx = None  #: int, None: index of the Operation in the circuit queue, or None if not in a queue
        if QNode._current_context is None:
            do_queue = False

        if self.num_wires == All:
            if do_queue:
                if set(wires) != set(range(QNode._current_context.num_wires)):
                    raise ValueError("Operator {} must act on all wires".format(self.name))

        if wires is None:
            raise ValueError("Must specify the wires that {} acts on".format(self.name))

        # extract the arguments
        params = args

        if len(params) != self.num_params:
            raise ValueError("{}: wrong number of parameters. "
                             "{} parameters passed, {} expected.".format(self.name, params, self.num_params))

        # check the validity of the params
        for p in params:
            self.check_domain(p)
        self.params = list(params)  #: list[Any]: parameters of the operation

        # apply the operator on the given wires
        if not isinstance(wires, Sequence):
            wires = [wires]
        self._wires = wires  #: Sequence[int, Variable]: wires on which the operation acts

        if all([isinstance(w, int) for w in self._wires]):
            # If all wires are integers (i.e., not Variable), check
            # that they are valid for the given operator
            self.check_wires(self._wires)

        if do_queue:
            self.queue()

    def __str__(self):
        """Print the operator name and some information."""
        return self.name +': {} params, wires {}'.format(len(self.params), self.wires)

    def check_wires(self, wires):
        """Check the validity of the operator wires.

        Args:
            wires (Sequence[int, Variable]): wires to check
        Raises:
            ValueError: list of wires is invalid
        Returns:
            Number, array, Variable: p
        """
        if self.num_wires != All and self.num_wires != Any and len(wires) != self.num_wires:
            raise ValueError("{}: wrong number of wires. "
                             "{} wires given, {} expected.".format(self.name, len(wires), self.num_wires))

        if len(set(wires)) != len(wires):
            raise ValueError('{}: wires must be unique, got {}.'.format(self.name, wires))

        return wires

    def check_domain(self, p, flattened=False):
        """Check the validity of a parameter.

        :class:`Variable` instances can represent any real scalars (but not arrays).

        Args:
            p (Number, array, Variable): parameter to check
            flattened (bool): True means p is an element of a flattened parameter
                sequence (affects the handling of 'A' parameters)
        Raises:
            TypeError: parameter is not an element of the expected domain
        Returns:
            Number, array, Variable: p
        """
        if isinstance(p, Variable):
            if self.par_domain == 'A':
                raise TypeError('{}: Array parameter expected, got a Variable, which can only represent real scalars.'.format(self.name))
            return p

        # p is not a Variable
        if self.par_domain == 'A':
            if flattened:
                if isinstance(p, np.ndarray):
                    raise TypeError('{}: Flattened array parameter expected, got {}.'.format(self.name, type(p)))
            else:
                if not isinstance(p, np.ndarray):
                    raise TypeError('{}: Array parameter expected, got {}.'.format(self.name, type(p)))
        elif self.par_domain in ('R', 'N'):
            if not isinstance(p, numbers.Real):
                raise TypeError('{}: Real scalar parameter expected, got {}.'.format(self.name, type(p)))

            if self.par_domain == 'N':
                if not isinstance(p, numbers.Integral):
                    raise TypeError('{}: Natural number parameter expected, got {}.'.format(self.name, type(p)))
                if p < 0:
                    raise TypeError('{}: Natural number parameter expected, got {}.'.format(self.name, p))
        else:
            raise ValueError('{}: Unknown parameter domain \'{}\'.'.format(self.name, self.par_domain))
        return p

    @property
    def wires(self):
        """Current wire values.

        Fixed wires are returned as is, free wires represented by
        :class:`~.variable.Variable` instances are replaced by their
        current numerical value.

        Returns:
            list[int]: wire values
        """
        w = [i.val if isinstance(i, Variable) else i for i in self._wires]
        self.check_wires(w)
        return [int(i) for i in w]

    @property
    def parameters(self):
        """Current parameter values.

        Fixed parameters are returned as is, free parameters represented by
        :class:`~.variable.Variable` instances are replaced by their
        current numerical value.

        Returns:
            list[float]: parameter values
        """
        temp = list(_flatten(self.params))
        temp_val = [self.check_domain(x.val, True) if isinstance(x, Variable) else x for x in temp]
        return _unflatten(temp_val, self.params)[0]

    def queue(self):
        """Append the operation to a QNode queue."""

        QNode._current_context._append_op(self)
        return self  # so pre-constructed Observable instances can be queued and returned in a single statement

#=============================================================================
# Base Operation class
#=============================================================================

class Operation(Operator):
    r"""Base class for quantum operations supported by a device.

    As with :class:`~.Operator`, the following class attributes must be
    defined for all operations:

    * :attr:`~.Operator.num_params`
    * :attr:`~.Operator.num_wires`
    * :attr:`~.Operator.par_domain`

    The following two class attributes are optional, but in most cases
    should be clearly defined to avoid unexpected behavior during
    differentiation.

    * :attr:`~.Operation.grad_method`
    * :attr:`~.Operation.grad_recipe`

    Finally, there are some additional optional class attributes
    that may be set, and used by certain quantum optimizers:

    * :attr:`~.Operation.generator`

    Args:
        args (tuple[float, int, array, Variable]): operation parameters

    Keyword Args:
        wires (Sequence[int]): Subsystems it acts on. If not given, args[-1]
            is interpreted as wires.
        do_queue (bool): Indicates whether the operation should be
            immediately pushed into a :class:`QNode` circuit queue.
            This flag is useful if there is some reason to run an Operation
            outside of a QNode context.
    """
    # pylint: disable=abstract-method
    _grad_recipe = None

    @property
    def grad_method(self):
        """Gradient computation method.

        * ``'A'``: analytic differentiation.
        * ``'F'``: finite difference numerical differentiation.
        * ``None``: the operation may not be differentiated.

        Default is ``'F'``, or ``None`` if the Operation has zero parameters.
        """
        return None if self.num_params == 0 else 'F'

    @property
    def grad_recipe(self):
        r"""Gradient recipe for the analytic differentiation method.

        This is a list with one tuple per operation parameter. For parameter
        :math:`k`, the tuple is of the form :math:`(c_k, s_k)`, resulting in
        a gradient recipe of

        .. math:: \frac{\partial}{\partial\phi_k}O = c_k\left[O(\phi_k+s_k)-O(\phi_k-s_k)\right].

        If this property returns ``None``, the default gradient recipe
        :math:`(c_k, s_k)=(1/2, \pi/2)` is assumed for every parameter.
        """
        return self._grad_recipe

    @property
    def generator(self):
        r"""Generator of the operation.

        A length-2 list ``[generator, scaling_factor]``, where

        * ``generator`` is an existing PennyLane
          operation class or :math:`2\times 2` Hermitian array
          that acts as the generator of the current operation

        * ``scaling_factor`` represents a scaling factor applied
          to the generator operation

        For example, if :math:`U(\theta)=e^{i0.7\theta \sigma_x}`, then
        :math:`\sigma_x`, with scaling factor :math:`s`, is the generator
        of operator :math:`U(\theta)`:

        .. code-block:: python

            generator = [PauliX, 0.7]

        Default is ``[None, 1]``, indicating the operation has no generator.
        """
        return [None, 1]

    @grad_recipe.setter
    def grad_recipe(self, value):
        """Setter for the grad_recipe property"""
        self._grad_recipe = value

    def __init__(self, *args, wires=None, do_queue=True):

        # check the grad_method validity
        if self.par_domain == 'N':
            assert self.grad_method is None, 'An operation may only be differentiated with respect to real scalar parameters.'
        elif self.par_domain == 'A':
            assert self.grad_method in (None, 'F'), 'Operations that depend on arrays containing free variables may only be differentiated using the F method.'

        # check the grad_recipe validity
        if self.grad_method == 'A':
            if self.grad_recipe is None:
                # default recipe for every parameter
                self.grad_recipe = [None] * self.num_params
            else:
                assert len(self.grad_recipe) == self.num_params, 'Gradient recipe must have one entry for each parameter!'
        else:
            assert self.grad_recipe is None, 'Gradient recipe is only used by the A method!'

        super().__init__(*args, wires=wires, do_queue=do_queue)


#=============================================================================
# Base Observable class
#=============================================================================


class Observable(Operator):
    """Base class for observables supported by a device.

    :class:`Observable` is used to describe Hermitian quantum observables.

    As with :class:`~.Operator`, the following class attributes must be
    defined for all observables:

    * :attr:`~.Operator.num_params`
    * :attr:`~.Operator.num_wires`
    * :attr:`~.Operator.par_domain`

    Args:
        args (tuple[float, int, array, Variable]): observable parameters

    Keyword Args:
        wires (Sequence[int]): subsystems it acts on.
            Currently, only one subsystem is supported.
        do_queue (bool): Indicates whether the operation should be
            immediately pushed into a :class:`QNode` observable queue.
            The observable queue is determined by the presence of an
            applicable `QNode._current_context`. If no context is
            available, this argument is ignored.
    """
    # pylint: disable=abstract-method
    return_type = None

    def __init__(self, *args, wires=None, do_queue=True):
        # extract the arguments
        if wires is not None:
            params = args
        else:
            params = args[:-1]
            wires = args[-1]

        super().__init__(*params, wires=wires, do_queue=do_queue)

    def __matmul__(self, other):
        if isinstance(other, Tensor):
            return other.__rmatmul__(self)

        if isinstance(other, Observable):
            return Tensor(self, other)

        raise ValueError("Can only perform tensor products between observables.")


class Tensor(Observable):
    """Container class representing tensor products of observables.

    To create a tensor, simply initiate it like so:

    >>> T = Tensor(qml.PauliX(0), qml.Hermitian(A, [1, 2]))

    You can also create a tensor from other Tensors:

    >>> T = Tensor(T, qml.PauliZ(4))

    The ``@`` symbol can be used as a tensor product operation:

    >>> T = qml.PauliX(0) @ qml.Hadamard(2)
    """
    return_type = None
    tensor = True
    par_domain = None

    def __init__(self, *args): #pylint: disable=super-init-not-called
        self.obs = []

        for o in args:
            if isinstance(o, Tensor):
                self.obs.extend(o.obs)
            elif isinstance(o, Observable):
                self.obs.append(o)
            else:
                raise ValueError("Can only perform tensor products between observables.")

    def __str__(self):
        """Print the tensor product and some information."""
        return 'Tensor product {}: {} params, wires {}'.format([i.name for i in self.obs], len(self.params), self.wires)

    @property
    def name(self):
        """All constituent observable names making up the tensor product.

        Returns:
            list[str]: list containing all observable names
        """
        return [o.name for o in self.obs]

    @property
    def num_wires(self):
        """Number of wires the tensor product acts on.

        Returns:
            int: number of wires
        """
        return len(list(_flatten(self.wires)))

    @property
    def wires(self):
        """All wires in the system the tensor product acts on.

        Returns:
            list[list[Any]]: nested list containing the wires per observable
            in the tensor product
        """
        return [o.wires for o in self.obs]

    @property
    def params(self):
        """Raw parameters of all constituent observables in the tensor product.

        Returns:
            list[Any]: flattened list containing all dependent parameters
        """
        return [p for sublist in [o.params for o in self.obs] for p in sublist]

    @property
    def num_params(self):
        """Raw parameters of all constituent observables in the tensor product.

        Returns:
            list[Any]: flattened list containing all dependent parameters
        """
        return len(self.params)

    @property
    def parameters(self):
        """Evaluated parameter values of all constituent observables in the tensor product.

        Returns:
            list[list[Any]]: nested list containing the parameters per observable
            in the tensor product
        """
        return [o.parameters for o in self.obs]

    def __matmul__(self, other):
        if isinstance(other, Tensor):
            self.obs.extend(other.obs)
            return self

        if isinstance(other, Observable):
            self.obs.append(other)
            return self

        raise ValueError("Can only perform tensor products between observables.")

    def __rmatmul__(self, other):
        if isinstance(other, Observable):
            self.obs[:0] = [other]
            return self

        raise ValueError("Can only perform tensor products between observables.")

    __imatmul__ = __matmul__


#=============================================================================
# CV Operations and observables
#=============================================================================


class CV:
    """A mixin base class denoting a continuous-variable operation."""
    # pylint: disable=no-member

    def heisenberg_expand(self, U, num_wires):
        """Expand the given local Heisenberg-picture array into a full-system one.

        Args:
            U (array[float]): array to expand (expected to be of the dimension ``1+2*self.num_wires``)
            num_wires (int): total number of wires in the quantum circuit. If zero, return ``U`` as is.
        Returns:
            array[float]: expanded array, dimension ``1+2*num_wires``
        """
        U_dim = len(U)
        nw = len(self.wires)


        if U.ndim > 2:
            raise ValueError('Only order-1 and order-2 arrays supported.')

        if U_dim != 1+2*nw:
            raise ValueError('{}: Heisenberg matrix is the wrong size {}.'.format(self.name, U_dim))

        if num_wires == 0 or list(self.wires) == list(range(num_wires)):
            # no expansion necessary (U is a full-system matrix in the correct order)
            return U

        if num_wires < len(self.wires):
            raise ValueError('{}: Number of wires {} is too small to fit Heisenberg matrix'.format(self.name, num_wires))

        # expand U into the I, x_0, p_0, x_1, p_1, ... basis
        dim = 1 + num_wires*2
        def loc(w):
            "Returns the slice denoting the location of (x_w, p_w) in the basis."
            ind = 2*w+1
            return slice(ind, ind+2)

        if U.ndim == 1:
            W = np.zeros(dim)
            W[0] = U[0]
            for k, w in enumerate(self.wires):
                W[loc(w)] = U[loc(k)]
        elif U.ndim == 2:
            if isinstance(self, Observable):
                W = np.zeros((dim, dim))
            else:
                W = np.eye(dim)

            W[0, 0] = U[0, 0]

            for k1, w1 in enumerate(self.wires):
                s1 = loc(k1)
                d1 = loc(w1)

                # first column
                W[d1, 0] = U[s1, 0]
                # first row (for gates, the first row is always (1, 0, 0, ...), but not for observables!)
                W[0, d1] = U[0, s1]

                for k2, w2 in enumerate(self.wires):
                    W[d1, loc(w2)] = U[s1, loc(k2)]  # block k1, k2 in U goes to w1, w2 in W.
        return W

    @staticmethod
    def _heisenberg_rep(p):
        r"""Heisenberg picture representation of the operation.

        * For Gaussian CV gates, this method returns the matrix of the linear
          transformation carried out by the gate for the given parameter values.
          The method is not defined for non-Gaussian gates.

          **The existence of this method is equivalent to setting** ``grad_method = 'A'``.

        * For observables, returns a real vector (first-order observables) or
          symmetric matrix (second-order observables) of expansion coefficients
          of the observable.

        For single-mode Operations we use the basis :math:`\mathbf{r} = (\I, \x, \p)`.
        For multi-mode Operations we use the basis :math:`\mathbf{r} = (\I, \x_0, \p_0, \x_1, \p_1, \ldots)`.

        .. note::

            For gates, we assume that the inverse transformation is obtained
            by negating the first parameter.

        Args:
            p (Sequence[float]): parameter values for the transformation

        Returns:
            array[float]: :math:`\tilde{U}` or :math:`q`
        """
        # pylint: disable=unused-argument
        return None

    @classproperty
    def supports_heisenberg(self):
        """Returns True if the CV Operation has
        overwritten the :meth:`~.CV._heisenberg_rep` static method
        defined in :class:`CV`, thereby indicating
        that analytic differentiation is supported if this operation
        succeeds the gate to be differentiated analytically.
        """
        return CV._heisenberg_rep != self._heisenberg_rep


class CVOperation(CV, Operation):
    """Base class for continuous-variable quantum operations."""
    # pylint: disable=abstract-method

    @classproperty
    def supports_analytic(self):
        """Returns True if the CV Operation has ``grad_method='A'`` and
        a defined :meth:`~.CV._heisenberg_rep` static method, indicating
        that analytic differentiation is supported.
        """
        return self.grad_method == 'A' and self.supports_heisenberg

    def heisenberg_pd(self, idx):
        """Partial derivative of the Heisenberg picture transform matrix.

        Computed using grad_recipe.

        Args:
            idx (int): index of the parameter with respect to which the
                partial derivative is computed.
        Returns:
            array[float]: partial derivative
        """
        # get the gradient recipe for this parameter
        recipe = self.grad_recipe[idx]
        multiplier = 0.5 if recipe is None else recipe[0]
        shift = np.pi / 2 if recipe is None else recipe[1]

        p = self.parameters
        # evaluate the transform at the shifted parameter values
        p[idx] += shift
        U2 = self._heisenberg_rep(p) # pylint: disable=assignment-from-none
        p[idx] -= 2*shift
        U1 = self._heisenberg_rep(p) # pylint: disable=assignment-from-none
        return (U2-U1) * multiplier  # partial derivative of the transformation

    def heisenberg_tr(self, num_wires, inverse=False):
        r"""Heisenberg picture representation of the linear transformation carried
        out by the gate at current parameter values.

        Given a unitary quantum gate :math:`U`, we may consider its linear
        transformation in the Heisenberg picture, :math:`U^\dagger(\cdot) U`.

        If the gate is Gaussian, this linear transformation preserves the polynomial order
        of any observables that are polynomials in :math:`\mathbf{r} = (\I, \x_0, \p_0, \x_1, \p_1, \ldots)`.
        This also means it maps :math:`\text{span}(\mathbf{r})` into itself:

        .. math:: U^\dagger \mathbf{r}_i U = \sum_j \tilde{U}_{ij} \mathbf{r}_j

        For Gaussian CV gates, this method returns the transformation matrix for
        the current parameter values of the Operation. The method is not defined
        for non-Gaussian (and non-CV) gates.

        Args:
            num_wires (int): total number of wires in the quantum circuit
            inverse  (bool): if True, return the inverse transformation instead

        Returns:
            array[float]: :math:`\tilde{U}`, the Heisenberg picture representation of the linear transformation
        """
        # not defined?
        p = self.parameters

        if self._heisenberg_rep(p) is None:
            raise RuntimeError('{} is not a Gaussian operation, or is missing the _heisenberg_rep method.'.format(self.name))

        if inverse:
            if self.par_domain == 'A':
                # TODO: expand this for the new par domain class, for non-unitary matrices.
                p[0] = np.linalg.inv(p[0])
            else:
                p[0] = -p[0]  # negate first parameter
        U = self._heisenberg_rep(p) # pylint: disable=assignment-from-none

        return self.heisenberg_expand(U, num_wires)


class CVObservable(CV, Observable):
    r"""Base class for continuous-variable observables.

    The class attribute :attr:`~.ev_order` can be defined to indicate
    to PennyLane whether the corresponding CV observable is a polynomial in the
    quadrature operators. If so,

    * ``ev_order = 1`` indicates a first order polynomial in quadrature
      operators :math:`(\x, \p)`.

    * ``ev_order = 2`` indicates a second order polynomial in quadrature
      operators :math:`(\x, \p)`.

    If :attr:`~.ev_order` is not ``None``, then the Heisenberg representation
    of the observable should be defined in the static method :meth:`~.CV._heisenberg_rep`,
    returning an array of the correct dimension.
    """
    # pylint: disable=abstract-method
    ev_order = None  #: None, int: if not None, the observable is a polynomial of the given order in `(x, p)`.

    def heisenberg_obs(self, num_wires):
        r"""Representation of the observable in the position/momentum operator basis.

        Returns the expansion :math:`q` of the observable, :math:`Q`, in the
        basis :math:`\mathbf{r} = (\I, \x_0, \p_0, \x_1, \p_1, \ldots)`.

        * For first-order observables returns a real vector such
          that :math:`Q = \sum_i q_i \mathbf{r}_i`.

        * For second-order observables returns a real symmetric matrix
          such that :math:`Q = \sum_{ij} q_{ij} \mathbf{r}_i \mathbf{r}_j`.

        Args:
            num_wires (int): total number of wires in the quantum circuit
        Returns:
            array[float]: :math:`q`
        """
        p = self.parameters
        U = self._heisenberg_rep(p) # pylint: disable=assignment-from-none
        return self.heisenberg_expand(U, num_wires)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains utilities and auxiliary functions which are shared
across the PennyLane submodules.
"""
from collections.abc import Iterable
import numbers
import inspect
import itertools

import numpy as np

from pennylane.variable import Variable


def _flatten(x):
    """Iterate through an arbitrarily nested structure, flattening it in depth-first order.

    See also :func:`_unflatten`.

    Args:
        x (array, Iterable, other): each element of the Iterable may itself be an iterable object

    Yields:
        other: elements of x in depth-first order
    """
    if isinstance(x, np.ndarray):
        yield from _flatten(x.flat)  # should we allow object arrays? or just "yield from x.flat"?
    elif isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
        for item in x:
            yield from _flatten(item)
    else:
        yield x


def _unflatten(flat, model):
    """Restores an arbitrary nested structure to a flattened iterable.

    See also :func:`_flatten`.

    Args:
        flat (array): 1D array of items
        model (array, Iterable, Number): model nested structure

    Returns:
        (other, array): first elements of flat arranged into the nested
        structure of model, unused elements of flat
    """
    if isinstance(model, (numbers.Number, Variable, str)):
        return flat[0], flat[1:]
    elif isinstance(model, np.ndarray):
        idx = model.size
        res = np.array(flat)[:idx].reshape(model.shape)
        return res, flat[idx:]
    elif isinstance(model, Iterable):
        res = []
        for x in model:
            val, flat = _unflatten(flat, x)
            res.append(val)
        return res, flat
    else:
        raise TypeError("Unsupported type in the model: {}".format(type(model)))


def unflatten(flat, model):
    """Wrapper for :func:`_unflatten`.
    """
    # pylint:disable=len-as-condition
    res, tail = _unflatten(np.asarray(flat), model)
    if len(tail) != 0:
        raise ValueError("Flattened iterable has more elements than the model.")
    return res


def _inv_dict(d):
    """Reverse a dictionary mapping.

    Returns multimap where the keys are the former values,
    and values are sets of the former keys.

    Args:
        d (dict[a->b]): mapping to reverse

    Returns:
        dict[b->set[a]]: reversed mapping
    """
    ret = {}
    for k, v in d.items():
        ret.setdefault(v, set()).add(k)
    return ret


def _get_default_args(func):
    """Get the default arguments of a function.

    Args:
        func (callable): a function

    Returns:
        dict[str, tuple]: mapping from argument name to (positional idx, default value)
    """
    signature = inspect.signature(func)
    return {
        k: (idx, v.default)
        for idx, (k, v) in enumerate(signature.parameters.items())
        if v.default is not inspect.Parameter.empty
    }


def expand(U, wires, num_wires):
    r"""Expand a multi-qubit operator into a full system operator.

    Args:
        U (array): :math:`2^n \times 2^n` matrix where n = len(wires).
        wires (Sequence[int]): Target subsystems (order matters! the
            left-most Hilbert space is at index 0).

    Returns:
        array: :math:`2^N\times 2^N` matrix. The full system operator.
    """
    if num_wires == 1:
        # total number of wires is 1, simply return the matrix
        return U

    N = num_wires
    wires = np.asarray(wires)

    if np.any(wires < 0) or np.any(wires >= N) or len(set(wires)) != len(wires):
        raise ValueError("Invalid target subsystems provided in 'wires' argument.")

    if U.shape != (2 ** len(wires), 2 ** len(wires)):
        raise ValueError("Matrix parameter must be of size (2**len(wires), 2**len(wires))")

    # generate N qubit basis states via the cartesian product
    tuples = np.array(list(itertools.product([0, 1], repeat=N)))

    # wires not acted on by the operator
    inactive_wires = list(set(range(N)) - set(wires))

    # expand U to act on the entire system
    U = np.kron(U, np.identity(2 ** len(inactive_wires)))

    # move active wires to beginning of the list of wires
    rearranged_wires = np.array(list(wires) + inactive_wires)

    # convert to computational basis
    # i.e., converting the list of basis state bit strings into
    # a list of decimal numbers that correspond to the computational
    # basis state. For example, [0, 1, 0, 1, 1] = 2^3+2^1+2^0 = 11.
    perm = np.ravel_multi_index(tuples[:, rearranged_wires].T, [2] * N)

    # permute U to take into account rearranged wires
    return U[:, perm][perm]



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.templates.parameters` module.
"""
# pylint: disable=protected-access,cell-var-from-loop
import pytest
import numpy as np
from pennylane.init import (cvqnn_layer_uniform,
                            cvqnn_layers_uniform,
                            cvqnn_layer_normal,
                            cvqnn_layers_normal,
                            interferometer_uniform,
                            interferometer_normal,
                            random_layer_uniform,
                            random_layers_uniform,
                            random_layer_normal,
                            random_layers_normal,
                            strong_ent_layer_uniform,
                            strong_ent_layers_uniform,
                            strong_ent_layer_normal,
                            strong_ent_layers_normal)


@pytest.fixture(scope="module",
                params=[2, 3])
def n_subsystems(request):
    """Number of qubits or modes."""
    return request.param


@pytest.fixture(scope="module",
                params=[None, 2, 10])
def n_rots(request):
    """Number of rotations in random layer."""
    return request.param


class TestParsCVQNN:
    """Tests the pennylane.init module for a cv-quantum neural network."""

    def test_cvqnnlayers_uniform_dimensions(self, n_subsystems, n_layers):
        """Confirm that pennylane.init.cvqnn_layers_uniform()
         returns an array with the right dimensions."""
        a = (n_layers, n_subsystems)
        b = (n_layers, n_subsystems * (n_subsystems - 1) // 2)
        p = cvqnn_layers_uniform(n_wires=n_subsystems, n_layers=n_layers, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a, a, a, b, b, a, a, a, a]

    def test_cvqnnlayers_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.cvqnn_layers_uniform() lies outside of interval."""
        low = -2
        high = 1
        p = cvqnn_layers_uniform(n_layers=2, n_wires=10, low=low, high=high, seed=seed)
        p_uni = [p[i] for i in [0, 1, 2, 4, 5, 6, 7, 9]]
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p_uni])

    def test_cvqnnlayers_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.cvqnn_layers_uniform()."""
        p = cvqnn_layers_uniform(n_layers=2, n_wires=10, low=1, high=1, mean_active=1, std_active=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_cvqnnlayers_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.cvqnn_layers_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        mean_a = 0.5
        std_a = 2
        n_wires = 3
        n_layers = 2
        n_if = n_wires * (n_wires - 1) // 2
        p = cvqnn_layers_uniform(n_layers=n_layers, n_wires=n_wires, low=low, high=high,
                                mean_active=mean_a, std_active=std_a, seed=seed)
        np.random.seed(seed)
        theta_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
        phi_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
        varphi_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
        r = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        phi_r = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
        theta_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
        phi_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
        varphi_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
        a = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        phi_a = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
        k = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        p_target = [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]
        assert np.allclose(p, p_target, atol=tol, rtol=0.)

    def test_cvqnnlayer_uniform_dimensions(self, n_subsystems):
        """Confirm that pennylane.init.cvqnn_layer_uniform()
         returns an array with the right dimensions."""
        a = (n_subsystems, )
        b = (n_subsystems * (n_subsystems - 1) // 2, )
        p = cvqnn_layer_uniform(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a, a, a, b, b, a, a, a, a]

    def test_cvqnnlayer_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.cvqnn_layer_uniform() lies outside of interval."""
        low = -2
        high = 1
        p = cvqnn_layer_uniform(n_wires=10, low=low, high=high, seed=seed)
        p_uni = [p[i] for i in [0, 1, 2, 4, 5, 6, 7, 9]]
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p_uni])

    def test_cvqnnlayer_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.cvqnn_layer_uniform()."""
        p = cvqnn_layer_uniform(n_wires=10, low=1, high=1, mean_active=1, std_active=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_cvqnnlayer_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.cvqnn_layer_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        mean_a = 0.5
        std_a = 2
        n_wires = 3
        n_if = n_wires * (n_wires - 1) // 2
        p = cvqnn_layer_uniform(n_wires=n_wires, low=low, high=high, mean_active=mean_a, std_active=std_a, seed=seed)
        np.random.seed(seed)
        theta_1 = np.random.uniform(low=low, high=high, size=(n_if,))
        phi_1 = np.random.uniform(low=low, high=high, size=(n_if,))
        varphi_1 = np.random.uniform(low=low, high=high, size=(n_wires,))
        r = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        phi_r = np.random.uniform(low=low, high=high, size=(n_wires,))
        theta_2 = np.random.uniform(low=low, high=high, size=(n_if,))
        phi_2 = np.random.uniform(low=low, high=high, size=(n_if,))
        varphi_2 = np.random.uniform(low=low, high=high, size=(n_wires,))
        a = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        phi_a = np.random.uniform(low=low, high=high, size=(n_wires,))
        k = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        p_target = [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]
        assert np.allclose(p, p_target, atol=tol)

    def test_cvqnnlayers_normal_dimensions(self, n_subsystems, n_layers):
        """Confirm that pennylane.init.cvqnn_layers_normal()
         returns an array with the right dimensions."""
        a = (n_layers, n_subsystems)
        b = (n_layers, n_subsystems * (n_subsystems - 1) // 2)
        p = cvqnn_layers_normal(n_wires=n_subsystems, n_layers=n_layers, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a, a, a, b, b, a, a, a, a]

    def test_cvqnnlayers_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.cvqnn_layers_normal()."""
        n_layers = 3
        p = cvqnn_layers_normal(n_layers=n_layers, n_wires=10, mean=1, std=0, mean_active=1, std_active=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_cvqnnlayers_normal_range(self, seed, tol):
        """Confirm that pennylane.init.cvqnn_layers_normal() invokes the correct np.random sampling function."""
        mean = -0.5
        std = 1
        mean_a = 0.5
        std_a = 2
        n_wires = 3
        n_layers = 3
        n_if = n_wires * (n_wires - 1) // 2
        p = cvqnn_layers_normal(n_layers=n_layers, n_wires=n_wires, mean=mean, std=std, mean_active=mean_a, std_active=std_a, seed=seed)
        np.random.seed(seed)
        theta_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
        phi_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
        varphi_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
        r = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        phi_r = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
        theta_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
        phi_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
        varphi_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
        a = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        phi_a = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
        k = np.random.normal(loc=mean_a, scale=std_a, size=(n_layers, n_wires))
        p_target = [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]
        assert np.allclose(p, p_target, atol=tol, rtol=0.)

    def test_cvqnnlayer_normal_dimensions(self, n_subsystems):
        """Confirm that pennylane.init.cvqnn_layer_normal()
         returns an array with the right dimensions."""
        a = (n_subsystems, )
        b = (n_subsystems * (n_subsystems - 1) // 2, )
        p = cvqnn_layer_normal(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a, a, a, b, b, a, a, a, a]

    def test_cvqnnlayer_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.cvqnn_layer_normal()."""
        p = cvqnn_layer_normal(n_wires=10, mean=1, std=0, mean_active=1, std_active=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_cvqnnlayer_normal_range(self, seed, tol):
        """Confirm that pennylane.init.cvqnn_layer_normal() invokes the correct np.random sampling function."""
        mean = -0.5
        std = 1
        mean_a = 0.5
        std_a = 2
        n_wires = 3
        n_if = n_wires * (n_wires - 1) // 2
        p = cvqnn_layer_normal(n_wires=n_wires, mean=mean, std=std, mean_active=mean_a, std_active=std_a, seed=seed)
        np.random.seed(seed)
        theta_1 = np.random.normal(loc=mean, scale=std, size=(n_if,))
        phi_1 = np.random.normal(loc=mean, scale=std, size=(n_if,))
        varphi_1 = np.random.normal(loc=mean, scale=std, size=(n_wires,))
        r = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        phi_r = np.random.normal(loc=mean, scale=std, size=(n_wires,))
        theta_2 = np.random.normal(loc=mean, scale=std, size=(n_if,))
        phi_2 = np.random.normal(loc=mean, scale=std, size=(n_if,))
        varphi_2 = np.random.normal(loc=mean, scale=std, size=(n_wires,))
        a = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        phi_a = np.random.normal(loc=mean, scale=std, size=(n_wires,))
        k = np.random.normal(loc=mean_a, scale=std_a, size=(n_wires,))
        p_target = [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]
        assert np.allclose(p, p_target, atol=tol)


class TestParsInterferometer:
    """Tests the pennylane.init module for an interferometer."""

    def test_interferometer_uniform_dimensions(self, n_subsystems):
        """Confirm that pennylane.init.interferometer_uniform()
         returns an array with the right dimensions."""
        a = (n_subsystems, )
        b = (n_subsystems * (n_subsystems - 1) // 2, )
        p = interferometer_uniform(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a]

    def test_interferometer_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.interferometer_uniform() lies outside of interval."""
        low = -2
        high = 1
        p = interferometer_uniform(n_wires=10, low=low, high=high, seed=seed)
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p])

    def test_interferometer_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.interferometer_uniform()."""
        p = interferometer_uniform(n_wires=10, low=1, high=1, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_interferometer_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.interferometer_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        n_wires = 3
        n_if = n_wires * (n_wires - 1) // 2
        p = interferometer_uniform(n_wires=n_wires, low=low, high=high, seed=seed)
        np.random.seed(seed)
        theta = np.random.uniform(low=low, high=high, size=(n_if,))
        phi = np.random.uniform(low=low, high=high, size=(n_if,))
        varphi = np.random.uniform(low=low, high=high, size=(n_wires,))
        p_target = [theta, phi, varphi]
        assert np.allclose(p, p_target, atol=tol, rtol=0.)

    def test_interferometer_normal_dimensions(self, n_subsystems):
        """Confirm that pennylane.init.interferometer_normal()
         returns an array with the right dimensions."""
        a = (n_subsystems, )
        b = (n_subsystems * (n_subsystems - 1) // 2, )
        p = interferometer_normal(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [b, b, a]

    def test_interferometer_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.interferometer_normal()."""
        p = interferometer_normal(n_wires=10, mean=1, std=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_interferometer_normal_seed(self, seed, tol):
        """Confirm that pennylane.init.interferometer_normal() invokes the correct np.random sampling function
        for a given seed."""
        mean = 0.5
        std = 1
        n_wires = 3
        n_if = n_wires * (n_wires - 1) // 2
        p = interferometer_normal(n_wires=n_wires, mean=mean, std=std, seed=seed)
        np.random.seed(seed)
        theta = np.random.normal(loc=mean, scale=std, size=(n_if,))
        phi = np.random.normal(loc=mean, scale=std, size=(n_if,))
        varphi = np.random.normal(loc=mean, scale=std, size=(n_wires,))
        p_target = [theta, phi, varphi]
        assert np.allclose(p, p_target, atol=tol, rtol=0.)


class TestParsStronglyEntangling:
    """Tests the pennylane.init module for a strongly entangling circuit."""

    def test_stronglyentanglinglayers_uniform_dimensions(self, n_subsystems, n_layers):
        """Confirm that the pennylane.init.strong_ent_layers_uniform()
         returns an array with the right dimensions."""
        a = (n_layers, n_subsystems, 3)
        p = strong_ent_layers_uniform(n_layers=n_layers, n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_stronglyentanglinglayers_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.strong_ent_layers_uniform() lies outside of interval."""
        low = -2
        high = 1
        n_layers = 3
        p = strong_ent_layers_uniform(n_layers=n_layers, n_wires=10, low=low, high=high, seed=seed)
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p])

    def test_stronglyentanglinglayers_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.strong_ent_layers_uniform()."""
        n_layers = 3
        p = strong_ent_layers_uniform(n_layers=n_layers, n_wires=10, low=1, high=1, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_stronglyentanglinglayers_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.strong_ent_layers_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        n_wires = 3
        n_layers = 3
        p = strong_ent_layers_uniform(n_layers=n_layers, n_wires=n_wires, low=low, high=high, seed=seed)
        np.random.seed(seed)
        p_target = np.random.uniform(low=low, high=high, size=(n_layers, n_wires, 3))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_stronglyentanglinglayer_uniform_dimensions(self, n_subsystems):
        """Confirm that the pennylane.init.strong_ent_layer_uniform()
         returns an array with the right dimensions."""
        a = (n_subsystems, 3)
        p = strong_ent_layer_uniform(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_stronglyentanglinglayer_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.strong_ent_layer_uniform() lies outside of interval."""
        low = -2
        high = 1
        p = strong_ent_layer_uniform(n_wires=10, low=low, high=high, seed=seed)
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p])

    def test_stronglyentanglinglayer_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.strong_ent_layer_uniform()."""
        p = strong_ent_layer_uniform(n_wires=10, low=1, high=1, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_stronglyentanglinglayer_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.strong_ent_layer_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        n_wires = 3
        p = strong_ent_layer_uniform(n_wires=n_wires, low=low, high=high, seed=seed)
        np.random.seed(seed)
        p_target = np.random.uniform(low=low, high=high, size=(n_wires, 3))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_stronglyentanglinglayers_normal_dimensions(self, n_subsystems, n_layers):
        """Confirm that the pennylane.init.strong_ent_layers_normal()
         returns an array with the right dimensions."""
        a = (n_layers, n_subsystems, 3)
        p = strong_ent_layers_normal(n_layers=n_layers, n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_stronglyentanglinglayers_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.strong_ent_layers_normal()."""
        n_layers = 3
        p = strong_ent_layers_normal(n_layers=n_layers, n_wires=10, mean=1, std=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_stronglyentanglinglayers_normal_seed(self, seed, tol):
        """Confirm that pennylane.init.strong_ent_layers_normal() invokes the correct np.random sampling function
        for a given seed."""
        mean = -2
        std = 1
        n_wires = 3
        n_layers = 3
        p = strong_ent_layers_normal(n_layers=n_layers, n_wires=n_wires, mean=mean, std=std, seed=seed)
        np.random.seed(seed)
        p_target = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires, 3))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_stronglyentanglinglayer_normal_dimensions(self, n_subsystems):
        """Confirm that the pennylane.init.parameters_stronglyentanglinglayer_normalm()
         returns an array with the right dimensions."""
        a = (n_subsystems, 3)
        p = strong_ent_layer_normal(n_wires=n_subsystems, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_stronglyentanglinglayer_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.strong_ent_layer_normal()."""
        p = strong_ent_layer_normal(n_wires=10, mean=1, std=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_stronglyentanglinglayer_normal_seed(self, seed, tol):
        """Confirm that pennylane.init.strong_ent_layer_normal() invokes the correct np.random sampling function
        for a given seed."""
        mean = -2
        std = 1
        n_wires = 3
        p = strong_ent_layer_normal(n_wires=n_wires, mean=mean, std=std, seed=seed)
        np.random.seed(seed)
        p_target = np.random.normal(loc=mean, scale=std, size=(n_wires, 3))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)


class TestParsRandom:
    """Tests the pennylane.init module for a random circuit."""

    def test_randomlayers_uniform_dimensions(self, n_subsystems, n_layers, n_rots):
        """Confirm that the pennylane.init.random_layers_uniform()
         returns an array with the right dimensions."""
        if n_rots is None:
            n_rots = n_subsystems
        a = (n_layers, n_rots)
        p = random_layers_uniform(n_layers=n_layers, n_wires=n_subsystems, n_rots=n_rots, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_randomlayers_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.random_layers_uniform() lies outside of interval."""
        low = -2
        high = 1
        n_layers = 3
        p = random_layers_uniform(n_layers=n_layers, n_wires=10, low=low, high=high, seed=seed)
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p])

    def test_randomlayers_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.random_layers_uniform()."""
        n_layers = 3
        p = random_layers_uniform(n_layers=n_layers, n_wires=10, low=1, high=1, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_randomlayers_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.random_layers_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        n_wires = 3
        n_rots = 5
        n_layers = 3
        p = random_layers_uniform(n_layers=n_layers, n_wires=n_wires, n_rots=n_rots, low=low, high=high, seed=seed)
        np.random.seed(seed)
        p_target = np.random.uniform(low=low, high=high, size=(n_layers, n_rots))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_randomlayer_uniform_dimensions(self, n_subsystems, n_rots):
        """Confirm that the pennylane.init.random_layer_uniform()
         returns an array with the right dimensions."""
        if n_rots is None:
            n_rots = n_subsystems
        a = (n_rots, )
        p = random_layer_uniform(n_wires=n_subsystems, n_rots=n_rots, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_randomlayer_uniform_interval(self, seed):
        """Confirm that no uniform sample in pennylane.init.random_layer_uniform() lies outside of interval."""
        low = -2
        high = 1
        p = random_layer_uniform(n_wires=10, low=low, high=high, seed=seed)
        assert all([(p_ <= high).all() and (p_ >= low).all() for p_ in p])

    def test_randomlayer_uniform_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.random_layer_uniform()."""
        p = random_layer_uniform(n_wires=10, low=1, high=1, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_randomlayer_uniform_seed(self, seed, tol):
        """Confirm that pennylane.init.random_layer_uniform() invokes the correct np.random sampling function
        for a given seed."""
        low = -2
        high = 1
        n_wires = 3
        n_rots = 5
        p = random_layer_uniform(n_wires=n_wires, n_rots=n_rots, low=low, high=high, seed=seed)
        np.random.seed(seed)
        p_target = np.random.uniform(low=low, high=high, size=(n_rots,))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_randomlayers_normal_dimensions(self, n_subsystems, n_layers, n_rots):
        """Confirm that the pennylane.init.random_layers_normal()
         returns an array with the right dimensions."""
        if n_rots is None:
            n_rots = n_subsystems
        a = (n_layers, n_rots)
        p = random_layers_normal(n_layers=n_layers, n_wires=n_subsystems, n_rots=n_rots, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_randomlayers_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.random_layers_normal()."""
        n_layers = 3
        p = random_layers_normal(n_layers=n_layers, n_wires=10, mean=1, std=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_randomlayers_normal_seed(self, seed, tol):
        """Confirm that pennylane.init.random_layers_normal() invokes the correct np.random sampling function
        for a given seed."""
        mean = -2
        std = 1
        n_wires = 3
        n_rots = 5
        n_layers = 3
        p = random_layers_normal(n_layers=n_layers, n_wires=n_wires, n_rots=n_rots, mean=mean, std=std, seed=seed)
        np.random.seed(seed)
        p_target = np.random.normal(loc=mean, scale=std, size=(n_layers, n_rots))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)

    def test_randomlayer_normal_dimensions(self, n_subsystems, n_rots):
        """Confirm that the pennylane.init.random_layer_normal()
         returns an array with the right dimensions."""
        if n_rots is None:
            n_rots = n_subsystems
        a = (n_rots, )
        p = random_layer_normal(n_wires=n_subsystems, n_rots=n_rots, seed=0)
        dims = [p_.shape for p_ in p]
        assert dims == [a]

    def test_randomlayer_normal_edgecase(self, seed, tol):
        """Test sampling edge case of pennylane.init.random_layer_normal()."""
        p = random_layer_normal(n_wires=10, mean=1, std=0, seed=seed)
        p_mean = np.mean(np.array([np.mean(pp) for p_ in p for pp in p_]))
        assert np.allclose(p_mean, 1, atol=tol, rtol=0.)

    def test_randomlayer_normal_seed(self, seed, tol):
        """Confirm that pennylane.init.random_layer_normal() invokes the correct np.random sampling function
        for a given seed."""
        mean = -2
        std = 1
        n_wires = 3
        n_rots = 5
        p = random_layer_normal(n_wires=n_wires, n_rots=n_rots, mean=mean, std=std, seed=seed)
        np.random.seed(seed)
        p_target = np.random.normal(loc=mean, scale=std, size=(n_rots,))
        assert np.allclose(p[0], p_target, atol=tol, rtol=0.)



r"""
.. _data_reuploading_classifier:

Data-reuploading classifer
==========================
*Author: Shahnawaz Ahmed (shahnawaz.ahmed95@gmail.com)*

A single-qubit quantum circuit which can implement arbitrary unitary
operations can be used as a universal classifier much like a single
hidden-layered Neural Network. As surprising as it sounds,
`Pérez-Salinas et al. (2019) <https://arxiv.org/abs/1907.02085>`_
discuss this with their idea of 'data
reuploading'. It is possible to load a single qubit with arbitrary
dimensional data and then use it as a universal classifier.

In this example, we will implement this idea with Pennylane - a
python based tool for quantum machine learning, automatic
differentiation, and optimization of hybrid quantum-classical
computations.

Background
----------

We consider a simple classification problem and will train a
single-qubit variational quantum circuit to achieve this goal. The data
is generated as a set of random points in a plane :math:`(x_1, x_2)` and
labeled as 1 (blue) or 0 (red) depending on whether they lie inside or
outside a circle. The goal is to train a quantum circuit to predict the
label (red or blue) given an input point’s coordinate.

.. figure:: ../../examples/figures/universal_circles.png
   :scale: 65%
   :alt: circles


Transforming quantum states using unitary operations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A single-qubit quantum state is characterized by a two-dimensional state
vector and can be visualized as a point in the so-called Bloch sphere.
Instead of just being a 0 (up) or 1 (down), it can exist in a
superposition with say 30% chance of being in the :math:`|0 \rangle` and
70% chance of being in the :math:`|1 \rangle` state. This is represented
by a state vector :math:`|\psi \rangle = 0.3|0 \rangle + 0.7|1 \rangle` -
the probability "amplitude" of the quantum state. In general we can take
a vector :math:`(\alpha, \beta)` to represent the probabilities of a qubit
being in a particular state and visualize it on the Bloch sphere as an
arrow.

.. figure:: ../../examples/figures/universal_bloch.png
   :scale: 65%
   :alt: bloch

Data loading using unitaries
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In order to load data onto a single qubit, we use a unitary operation
:math:`U(x_1, x_2, x_3)` which is just a parameterized
matrix multiplication representing the rotation of the state vector in
the Bloch sphere. E.g., to load :math:`(x_1, x_2)` into the qubit, we
just start from some initial state vector, :math:`|0 \rangle`,
apply the unitary operation :math:`U(x_1, x_2, 0)` and end up at a new
point on the Bloch sphere. Here we have padded 0 since our data is only
2D. Pérez-Salinas et al. (2019) discuss how to load a higher
dimensional data point (:math:`[x_1, x_2, x_3, x_4, x_5, x_6]`) by
breaking it down in sets of three parameters
(:math:`U(x_1, x_2, x_3), U(x_4, x_5, x_6)`).

Model parameters with data re-uploading
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Once we load the data onto the quantum circuit, we want to have some
trainable nonlinear model similar to a neural network as well as a way of
learning the weights of the model from data. This is again done with
unitaries, :math:`U(\theta_1, \theta_2, \theta_3)`, such that we load the
data first and then apply the weights to form a single layer
:math:`L(\vec \theta, \vec x) = U(\vec \theta)U(\vec x)`. In principle,
this is just application of two matrix multiplications on an input
vector initialized to some value. In order to increase the number of
trainable parameters (similar to increasing neurons in a single layer of
a neural network), we can reapply this layer again and again with new
sets of weights,
:math:`L(\vec \theta_1, \vec x) L(\vec \theta_2, , \vec x) ... L(\vec \theta_L, \vec x)`
for :math:`L` layers. The quantum circuit would look like the following:

.. figure:: ../../examples/figures/universal_layers.png
   :scale: 75%
   :alt: Layers


The cost function and "nonlinear collapse"
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

So far, we have only performed linear operations (matrix
multiplications) and we know that we need to have some nonlinear
squashing similar to activation functions in neural networks to really
make a universal classifier (Cybenko 1989). Here is where things gets a
bit quantum. After the application of the layers, we will end up at some
point on the Bloch sphere due to the sequence of unitaries implementing
rotations of the input. These are still just linear transformations of
the input state. Now, the output of the model should be a class label
which can be encoded as fixed vectors (Blue = :math:`[1, 0]`, Red =
:math:`[0, 1]`) on the Bloch sphere. We want to end up at either of them
after transforming our input state through alternate applications of
data layer and weights.

We can use the idea of the “collapse” of our quantum state into
one or other class. This happens when we measure the quantum state which
leads to its projection as either the state 0 or 1. We can compute the
fidelity (or closeness) of the output state to the class label making
the output state jump to either :math:`| 0 \rangle` or
:math:`|1\rangle`. By repeating this process several times, we can
compute the probability or overlap of our output to both labels and
assign a class based on the label our output has a higher overlap. This
is much like having a set of output neurons and selecting the one which
has the highest value as the label.

We can encode the output label as a particular quantum state that we want
to end up in and use Pennylane to find the probability of ending up in that
state after running the circuit. We construct an observable corresponding to
the output label using the `Hermitian <https://pennylane.readthedocs.io/en/latest/code/ops/qubit.html#pennylane.ops.qubit.Hermitian>`_
operator. The expectation value of the observable gives the overlap or fidelity.
We can then define the cost function as the sum of the fidelities for all
the data points after passing through the circuit and optimize the parameters
:math:`(\vec \theta)` to minimize the cost.

.. math::

   \texttt{Cost} = \sum_{\texttt{data points}} (1 - \texttt{fidelity}(\psi_{\texttt{output}}(\vec x, \vec \theta), \psi_{\texttt{label}}))

Now, we can use our favorite optimizer to maximize the sum of the
fidelities over all data points (or batches of datapoints) and find the
optimal weights for classification. Gradient-based optimizers such as
Adam (Kingma et. al., 2014) can be used if we have a good model of
the circuit and how noise might affect it. Or, we can use some
gradient-free method such as L-BFGS (Liu, Dong C., and Nocedal, J., 1989)
to evaluate the gradient and find the optimal weights where we can
treat the quantum circuit as a black-box and the gradients are computed
numerically using a fixed number of function evalutaions and iterations.
The L-BFGS method can be used with the PyTorch interface for Pennylane.

Multiple qubits, entanglement and Deep Neural Networks
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The Universal Approximation Theorem declares that a neural network with
two or more hidden layers can serve as a universal function approximator.
Recently, we have witnessed remarkable progress of learning algorithms using
Deep Neural Networks.

Pérez-Salinas et al. (2019) make a connection to Deep Neural Networks by
describing that in their approach the
“layers” :math:`L_i(\vec \theta_i, \vec x )` are analogous to the size
of the intermediate hidden layer of a neural network. And the concept of
deep (multiple layers of the neural network) relates to the number
of qubits. So, multiple qubits with entanglement between them could
provide some quantum advantage over classical neural networks. But here,
we will only implement a single qubit classifier.

.. figure:: ../../examples/figures/universal_dnn.png
   :scale: 35%
   :alt: DNN

"Talk is cheap. Show me the code." - Linus Torvalds
---------------------------------------------------
"""

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import AdamOptimizer, GradientDescentOptimizer

import matplotlib.pyplot as plt


# Set a random seed
np.random.seed(42)


# Make a dataset of points inside and outside of a circle
def circle(samples, center=[0.0, 0.0], radius=np.sqrt(2 / np.pi)):
    """
    Generates a dataset of points with 1/0 labels inside a given radius.

    Args:
        samples (int): number of samples to generate
        center (tuple): center of the circle
        radius (float: radius of the circle

    Returns:
        Xvals (array[tuple]): coordinates of points
        Xvals (array[int]): classification labels
    """
    Xvals, yvals = [], []

    for i in range(samples):
        x = 2 * (np.random.rand(2)) - 1
        y = 0
        if np.linalg.norm(x - center) < radius:
            y = 1
        Xvals.append(x)
        yvals.append(y)
    return np.array(Xvals), np.array(yvals)


def plot_data(x, y, fig=None, ax=None):
    """
    Plot data with red/blue values for a binary classification.

    Args:
        x (array[tuple]): array of data points as tuples
        y (array[int]): array of data points as tuples
    """
    if fig == None:
        fig, ax = plt.subplots(1, 1, figsize=(5, 5))
    reds = y == 0
    blues = y == 1
    ax.scatter(x[reds, 0], x[reds, 1], c="red", s=20, edgecolor="k")
    ax.scatter(x[blues, 0], x[blues, 1], c="blue", s=20, edgecolor="k")
    ax.set_xlabel("$x_1$")
    ax.set_ylabel("$x_2$")


Xdata, ydata = circle(500)
fig, ax = plt.subplots(1, 1, figsize=(4, 4))
plot_data(Xdata, ydata, fig=fig, ax=ax)
plt.show()


# Define output labels as quantum state vectors
def density_matrix(state):
    """Calculates the density matrix representation of a state.

    Args:
        state (array[complex]): array representing a quantum state vector

    Returns:
        dm: (array[complex]): array representing the density matrix
    """
    return state * np.conj(state).T


label_0 = [[1], [0]]
label_1 = [[0], [1]]
state_labels = [label_0, label_1]


##############################################################################
# Simple classifier with data reloading and fidelity loss
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

dev = qml.device("default.qubit", wires=1)
# Install any pennylane-plugin to run on some particular backend


@qml.qnode(dev)
def qcircuit(params, x=None, y=None):
    """A variational quantum circuit representing the Universal classifier.

    Args:
        params (array[float]): array of parameters
        x (array[float]): single input vector
        y (array[float]): single output state density matrix

    Returns:
        float: fidelity between output state and input
    """
    for p in params:
        qml.Rot(*x, wires=0)
        qml.Rot(*p, wires=0)
    return qml.expval(qml.Hermitian(y, wires=[0]))


def fidelity(state1, state2):
    """
    Calculates the fidelity between two state vectors

    Args:
        state1 (array[float]): State vector representation
        state2 (array[float]): State vector representation

    Returns:
        float: fidelity between `state1` and `state2`
    """
    return np.abs(np.dot(np.conj(state1), state2))


def cost(params, x, y, state_labels=None):
    """Cost function to be minimized.

    Args:
        params (array[float]): array of parameters
        x (array[float]): 2-d array of input vectors
        y (array[float]): 1-d array of targets
        state_labels (array[float]): array of state representations for labels

    Returns:
        float: loss value to be minimized
    """
    # Compute prediction for each input in data batch
    loss = 0.0
    dm_labels = [density_matrix(s) for s in state_labels]
    for i in range(len(x)):
        f = qcircuit(params, x=x[i], y=dm_labels[y[i]])
        loss = loss + (1 - f) ** 2
    return loss / len(x)


##############################################################################
# Utility functions for testing and creating batches
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def test(params, x, y, state_labels=None):
    """
    Tests on a given set of data.

    Args:
        params (array[float]): array of parameters
        x (array[float]): 2-d array of input vectors
        y (array[float]): 1-d array of targets
        state_labels (array[float]): 1-d array of state representations for labels

    Returns:
        predicted (array([int]): predicted labels for test data
        output_states (array[float]): output quantum states from the circuit
    """
    fidelity_values = []
    output_states = []
    dm_labels = [density_matrix(s) for s in state_labels]
    for i in range(len(x)):
        expectation = qcircuit(params, x=x[i], y=dm_labels[y[i]])
        output_states.append(dev._state)
    predicted = predicted_labels(output_states, state_labels)
    return predicted, output_states


def predicted_labels(states, state_labels=None):
    """
    Computes the label of the predicted state by selecting the one
    with maximum fidelity.

    Args:
        weights (array[float]): array of weights
        x (array[float]): 2-d array of input vectors
        y (array[float]): 1-d array of targets
        state_labels (array[float]): 1-d array of state representations for labels

    Returns:
        float: loss value to be minimized
    """
    output_labels = [
        np.argmax([fidelity(s, label) for label in state_labels]) for s in states
    ]
    return np.array(output_labels)


def accuracy_score(y_true, y_pred):
    """Accuracy score.

    Args:
        y_true (array[float]): 1-d array of targets
        y_predicted (array[float]): 1-d array of predictions
        state_labels (array[float]): 1-d array of state representations for labels

    Returns:
        score (float): the fraction of correctly classified samples
    """
    score = y_true == y_pred
    return score.sum() / len(y_true)


def iterate_minibatches(inputs, targets, batch_size):
    """
    A generator for batches of the input data

    Args:
        inputs (array[float]): input data
        targets (array[float]): targets

    Returns:
        inputs (array[float]): one batch of input data of length `batch_size`
        targets (array[float]): one batch of targets of length `batch_size`
    """
    for start_idx in range(0, inputs.shape[0] - batch_size + 1, batch_size):
        idxs = slice(start_idx, start_idx + batch_size)
        yield inputs[idxs], targets[idxs]


##############################################################################
# Train a quantum classifier on the circle dataset
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Generate training and test data
num_training = 200
num_test = 2000

Xdata, y_train = circle(num_training)
X_train = np.hstack((Xdata, np.zeros((Xdata.shape[0], 1))))

Xtest, y_test = circle(num_test)
X_test = np.hstack((Xtest, np.zeros((Xtest.shape[0], 1))))


# Train using Adam optimizer and evaluate the classifier
num_layers = 3
learning_rate = 0.6
epochs = 20
batch_size = 32

opt = AdamOptimizer(learning_rate, beta1=0.9, beta2=0.999)

# initialize random weights
params = np.random.uniform(size=(num_layers, 3))

predicted_train, states_train = test(params, X_train, y_train, state_labels)
accuracy_train = accuracy_score(y_train, predicted_train)

predicted_test, states_test = test(params, X_test, y_test, state_labels)
accuracy_test = accuracy_score(y_test, predicted_test)

# save predictions with random weights for comparison
initial_predictions = predicted_test

loss = cost(params, X_test, y_test, state_labels)

print(
    "Epoch: {:2d} | Cost: {:3f} | Train accuracy: {:3f} | Test Accuracy: {:3f}".format(
        0, loss, accuracy_train, accuracy_test))

for it in range(epochs):
    for Xbatch, ybatch in iterate_minibatches(X_train, y_train, batch_size=batch_size):
        params = opt.step(lambda v: cost(v, Xbatch, ybatch, state_labels), params)

    predicted_train, states_train = test(params, X_train, y_train, state_labels)
    accuracy_train = accuracy_score(y_train, predicted_train)
    loss = cost(params, X_train, y_train, state_labels)

    predicted_test, states_test = test(params, X_test, y_test, state_labels)
    accuracy_test = accuracy_score(y_test, predicted_test)
    res = [it + 1, loss, accuracy_train, accuracy_test]
    print(
        "Epoch: {:2d} | Loss: {:3f} | Train accuracy: {:3f} | Test accuracy: {:3f}".format(
            *res))


##############################################################################
# Results
# ~~~~~~~

print("Cost: {:3f} | Train accuracy {:3f} | Test Accuracy : {:3f}".format(
        loss, accuracy_train, accuracy_test))

print("Learned weights")
for i in range(num_layers):
    print("Layer {}: {}".format(i, params[i]))


fig, axes = plt.subplots(1, 3, figsize=(10, 3))
plot_data(X_test, initial_predictions, fig, axes[0])
plot_data(X_test, predicted_test, fig, axes[1])
plot_data(X_test, y_test, fig, axes[2])
axes[0].set_title("Predictions with random weights")
axes[1].set_title("Predictions after training")
axes[2].set_title("True test data")
plt.show()


##############################################################################
# This tutorial was generated using the following Pennylane version:

qml.about()


##############################################################################
# References
# ----------
# [1] Pérez-Salinas, Adrián, et al. “Data re-uploading for a universal
# quantum classifier.” arXiv preprint arXiv:1907.02085 (2019).
#
# [2] Kingma, Diederik P., and Ba, J. "Adam: A method for stochastic
# optimization." arXiv preprint arXiv:1412.6980 (2014).
#
# [3] Liu, Dong C., and Nocedal, J. "On the limited memory BFGS
# method for large scale optimization." Mathematical programming
# 45.1-3 (1989): 503-528.



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the :class:`Device` abstract base class.
"""
# pylint: disable=too-many-format-args
import abc

import numpy as np

from pennylane.operation import Operation, Observable, Sample, Variance, Expectation, Tensor
from .qnode import QuantumFunctionError


class DeviceError(Exception):
    """Exception raised by a :class:`~.pennylane._device.Device` when it encounters an illegal
    operation in the quantum circuit.
    """


class Device(abc.ABC):
    """Abstract base class for PennyLane devices.

    Args:
        wires (int): number of subsystems in the quantum state represented by the device.
            Default 1 if not specified.
        shots (int): Number of circuit evaluations/random samples used to estimate
            expectation values of observables. Defaults to 1000 if not specified.
    """
    #pylint: disable=too-many-public-methods
    _capabilities = {} #: dict[str->*]: plugin capabilities
    _circuits = {}     #: dict[str->Circuit]: circuit templates associated with this API class

    def __init__(self, wires=1, shots=1000):
        self.num_wires = wires
        self._shots = 1000
        self.shots = shots

        self._op_queue = None
        self._obs_queue = None
        self._parameters = None

    def __repr__(self):
        """String representation."""
        return "{}.\nInstance: ".format(self.__module__, self.__class__.__name__, self.name)

    def __str__(self):
        """Verbose string representation."""
        return "{}\nName: \nAPI version: \nPlugin version: \nAuthor: ".format(self.name, self.pennylane_requires, self.version, self.author)

    @property
    @abc.abstractmethod
    def name(self):
        """The full name of the device."""

    @property
    @abc.abstractmethod
    def short_name(self):
        """Returns the string used to load the device."""

    @property
    @abc.abstractmethod
    def pennylane_requires(self):
        """The current API version that the device plugin was made for."""

    @property
    @abc.abstractmethod
    def version(self):
        """The current version of the plugin."""

    @property
    @abc.abstractmethod
    def author(self):
        """The author(s) of the plugin."""

    @property
    @abc.abstractmethod
    def operations(self):
        """Get the supported set of operations.

        Returns:
            set[str]: the set of PennyLane operation names the device supports
        """

    @property
    @abc.abstractmethod
    def observables(self):
        """Get the supported set of observables.

        Returns:
            set[str]: the set of PennyLane observable names the device supports
        """

    @property
    def shots(self):
        """Number of circuit evaluations/random samples used to estimate
        expectation values of observables"""

        return self._shots

    @shots.setter
    def shots(self, shots):
        """Changes the number of shots.

        Args:
            shots (int): number of circuit evaluations/random samples used to estimate
                expectation values of observables
        """
        if shots < 1:
            raise DeviceError("The specified number of shots needs to be at least 1. Got {}.".format(shots))

        self._shots = int(shots)

    @classmethod
    def capabilities(cls):
        """Get the other capabilities of the plugin.

        Measurements, batching etc.

        Returns:
            dict[str->*]: results
        """
        return cls._capabilities

    def execute(self, queue, observables, parameters={}):
        """Execute a queue of quantum operations on the device and then measure the given observables.

        For plugin developers: Instead of overwriting this, consider implementing a suitable subset of
        :meth:`pre_apply`, :meth:`apply`, :meth:`post_apply`, :meth:`pre_measure`,
        :meth:`expval`, :meth:`var`, :meth:`sample`, :meth:`post_measure`, and :meth:`execution_context`.

        Args:
            queue (Iterable[~.operation.Operation]): operations to execute on the device
            observables (Iterable[~.operation.Observable]): observables to measure and return
            parameters (dict[int->list[(int, int)]]): Mapping from free parameter index to the list of
                :class:`Operations <pennylane.operation.Operation>` (in the queue) that depend on it.
                The first element of the tuple is the index of the Operation in the program queue,
                the second the index of the parameter within the Operation.

        Returns:
            array[float]: measured value(s)
        """
        self.check_validity(queue, observables)
        self._op_queue = queue
        self._obs_queue = observables
        self._parameters = {}
        self._parameters.update(parameters)

        results = []

        with self.execution_context():
            self.pre_apply()

            for operation in queue:
                self.apply(operation.name, operation.wires, operation.parameters)

            self.post_apply()

            self.pre_measure()

            for obs in observables:
                if obs.return_type is Expectation:
                    results.append(self.expval(obs.name, obs.wires, obs.parameters))

                elif obs.return_type is Variance:
                    results.append(self.var(obs.name, obs.wires, obs.parameters))

                elif obs.return_type is Sample:
                    results.append(np.array(self.sample(obs.name, obs.wires, obs.parameters)))

                elif obs.return_type is not None:
                    raise QuantumFunctionError("Unsupported return type specified for observable {}".format(obs.name))

            self.post_measure()

            self._op_queue = None
            self._obs_queue = None
            self._parameters = None

            # Ensures that a combination with sample does not put
            # expvals and vars in superfluous arrays
            if all(obs.return_type is Sample for obs in observables):
                return np.asarray(results)
            if any(obs.return_type is Sample for obs in observables):
                return np.asarray(results, dtype="object")

            return np.asarray(results)

    @property
    def op_queue(self):
        """The operation queue to be applied.

        Note that this property can only be accessed within the execution context
        of :meth:`~.execute`.

        Returns:
            list[~.operation.Operation]
        """
        if self._op_queue is None:
            raise ValueError("Cannot access the operation queue outside of the execution context!")

        return self._op_queue

    @property
    def obs_queue(self):
        """The observables to be measured and returned.

        Note that this property can only be accessed within the execution context
        of :meth:`~.execute`.

        Returns:
            list[~.operation.Observable]
        """
        if self._obs_queue is None:
            raise ValueError("Cannot access the observable value queue outside of the execution context!")

        return self._obs_queue

    @property
    def parameters(self):
        """Mapping from free parameter index to the list of
        :class:`Operations <~.Operation>` in the device queue that depend on it.

        Note that this property can only be accessed within the execution context
        of :meth:`~.execute`.

        Returns:
            dict[int->list[(int, int)]]: the first element of the tuple is the index
            of the Operation in the program queue, the second the index of the parameter
            within the Operation.
        """
        if self._parameters is None:
            raise ValueError("Cannot access the free parameter mapping outside of the execution context!")

        return self._parameters

    def pre_apply(self):
        """Called during :meth:`execute` before the individual operations are executed."""

    def post_apply(self):
        """Called during :meth:`execute` after the individual operations have been executed."""

    def pre_measure(self):
        """Called during :meth:`execute` before the individual observables are measured."""

    def post_measure(self):
        """Called during :meth:`execute` after the individual observables have been measured."""

    def execution_context(self):
        """The device execution context used during calls to :meth:`execute`.

        You can overwrite this function to return a context manager in case your
        quantum library requires that;
        all operations and method calls (including :meth:`apply` and :meth:`expval`)
        are then evaluated within the context of this context manager (see the
        source of :meth:`.Device.execute` for more details).
        """
        # pylint: disable=no-self-use
        class MockContext: # pylint: disable=too-few-public-methods
            """Mock class as a default for the with statement in execute()."""
            def __enter__(self):
                pass
            def __exit__(self, type, value, traceback):
                pass

        return MockContext()

    def supports_operation(self, operation):
        """Checks if an operation is supported by this device.

        Args:
            operation (Operation,str): operation to be checked

        Returns:
            bool: ``True`` iff supplied operation is supported
        """
        if isinstance(operation, type) and issubclass(operation, Operation):
            return operation.__name__ in self.operations
        if isinstance(operation, str):
            return operation in self.operations

        raise ValueError("The given operation must either be a pennylane.Operation class or a string.")

    def supports_observable(self, observable):
        """Checks if an observable is supported by this device.

        Args:
            operation (Observable,str): observable to be checked

        Returns:
            bool: ``True`` iff supplied observable is supported
        """
        if isinstance(observable, type) and issubclass(observable, Observable):
            return observable.__name__ in self.observables
        if isinstance(observable, str):
            return observable in self.observables

        raise ValueError("The given operation must either be a pennylane.Observable class or a string.")

    def check_validity(self, queue, observables):
        """Checks whether the operations and observables in queue are all supported by the device.

        Args:
            queue (Iterable[~.operation.Operation]): quantum operation objects which are intended
                to be applied on the device
            observables (Iterable[~.operation.Observable]): observables which are intended
                to be evaluated on the device
        """
        for o in queue:
            if o.name not in self.operations:
                raise DeviceError("Gate {} not supported on device {}".format(o.name, self.short_name))

        for o in observables:

            if isinstance(o, Tensor):
                if "tensor_observables" not in self.capabilities() or not self.capabilities()["tensor_observables"]:
                    raise DeviceError("Tensor observables not supported on device {}".format(self.short_name))

                for i in o.obs:
                    if i.name not in self.observables:
                        raise DeviceError("Observable {} not supported on device {}".format(i.name, self.short_name))
            else:
                if o.name not in self.observables:
                    raise DeviceError("Observable {} not supported on device {}".format(o.name, self.short_name))

    @abc.abstractmethod
    def apply(self, operation, wires, par):
        """Apply a quantum operation.

        For plugin developers: this function should apply the operation on the device.

        Args:
            operation (str): name of the operation
            wires (Sequence[int]): subsystems the operation is applied on
            par (tuple): parameters for the operation
        """

    @abc.abstractmethod
    def expval(self, observable, wires, par):
        r"""Returns the expectation value of observable on specified wires.

        Note: all arguments accept _lists_, which indicate a tensor
        product of observables.

        Args:
            observable (str or list[str]): name of the observable(s)
            wires (List[int] or List[List[int]]): subsystems the observable(s) is to be measured on
            par (tuple or list[tuple]]): parameters for the observable(s)

        Returns:
            float: expectation value :math:`\expect{A} = \bra{\psi}A\ket{\psi}`
        """

    def var(self, observable, wires, par):
        r"""Returns the variance of observable on specified wires.

        Note: all arguments support _lists_, which indicate a tensor
        product of observables.

        Args:
            observable (str or list[str]): name of the observable(s)
            wires (List[int] or List[List[int]]): subsystems the observable(s) is to be measured on
            par (tuple or list[tuple]]): parameters for the observable(s)

        Returns:
            float: variance :math:`\mathrm{var}(A) = \bra{\psi}A^2\ket{\psi} - \bra{\psi}A\ket{\psi}^2`
        """
        raise NotImplementedError("Returning variances from QNodes not currently supported by {}".format(self.short_name))

    def sample(self, observable, wires, par):
        """Return a sample of an observable.

        The number of samples is determined by the value of ``Device.shots``,
        which can be directly modified.

        Note: all arguments support _lists_, which indicate a tensor
        product of observables.

        Args:
            observable (str or list[str]): name of the observable(s)
            wires (List[int] or List[List[int]]): subsystems the observable(s) is to be measured on
            par (tuple or list[tuple]]): parameters for the observable(s)

        Returns:
            array[float]: samples in an array of dimension ``(n, num_wires)``
        """
        raise NotImplementedError("Returning samples from QNodes not currently supported by {}".format(self.short_name))

    def probability(self):
        """Return the full state probability of each computational basis state from the last run of the device.

        Returns:
            OrderedDict[tuple, float]: Dictionary mapping a tuple representing the state
            to the resulting probability. The dictionary should be sorted such that the
            state tuples are in lexicographical order.
        """
        raise NotImplementedError("Returning probability not currently supported by {}".format(self.short_name))

    @abc.abstractmethod
    def reset(self):
        """Reset the backend state.

        After the reset the backend should be as if it was just constructed.
        Most importantly the quantum state is reset to its initial value.
        """



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the :func:`qnode` decorator, which turns a quantum function into
a :class:`Qnode` object.
"""

# pylint: disable=redefined-outer-name
from functools import wraps, lru_cache

from .qnode import QNode


def qnode(device, interface='numpy', cache=False):
    """QNode decorator.

    Args:
        device (~pennylane._device.Device): a PennyLane-compatible device
        interface (str): the interface that will be used for automatic
            differentiation and classical processing. This affects
            the types of objects that can be passed to/returned from the QNode:

            * ``interface='numpy'``: The QNode accepts default Python types
              (floats, ints, lists) as well as NumPy array arguments,
              and returns NumPy arrays.

            * ``interface='torch'``: The QNode accepts and returns Torch tensors.

            * ``interface='tfe'``: The QNode accepts and returns eager execution
              TensorFlow ``tfe.Variable`` objects.

        cache (bool): If ``True``, the quantum function used to generate the QNode will
            only be called to construct the quantum circuit once, on first execution,
            and this circuit structure (i.e., the placement of templates, gates, measurements, etc.) will be cached for all further executions. The circuit parameters can still change with every call. Only activate this
            feature if your quantum circuit structure will never change.
    """
    @lru_cache()
    def qfunc_decorator(func):
        """The actual decorator"""

        qnode = QNode(func, device, cache=cache)

        if interface == 'torch':
            return qnode.to_torch()

        if interface == 'tf':
            return qnode.to_tf()

        @wraps(func)
        def wrapper(*args, **kwargs):
            """Wrapper function"""
            return qnode(*args, **kwargs)

        # bind the jacobian method to the wrapped function
        wrapper.jacobian = qnode.jacobian
        wrapper.metric_tensor = qnode.metric_tensor
        wrapper.print_applied = qnode.print_applied

        # bind the qnode attributes to the wrapped function
        wrapper.__dict__.update(qnode.__dict__)

        return wrapper
    return qfunc_decorator



r"""
.. _gaussian_transformation:

Gaussian transformation
=======================

This tutorial demonstrates the basic working principles of PennyLane for
continuous-variable (CV) photonic devices. For more details about photonic
quantum computing, the
`Strawberry Fields documentation <https://strawberryfields.readthedocs.io/en/latest/>`_
is a great starting point.

The quantum circuit
-------------------

For this basic tutorial, we will consider a special subset of CV operations:
the *Gaussian transformations*. We work with the following simple Gaussian circuit:

.. figure:: ../../examples/figures/gaussian_transformation.svg
    :align: center
    :width: 40%
    :target: javascript:void(0);

What is this circuit doing?

1. **We begin with one wire (qumode) in the vacuum state**. Note that we use the same
   notation :math:`|0\rangle` for the initial state as the previous qubit tutorial.
   In a photonic CV system, this state is the *vacuum state*, i.e., the state with no
   photons in the wire.

2. **We displace the qumode**. The displacement gate linearly shifts the state of the
   qumode in phase space. The vacuum state is centered at the origin in phase space,
   while the displaced state will be centered at the point :math:`\alpha`.

3. **We rotate the qumode**. This is another linear transformation in phase space,
   albeit a rotation (by angle :math:`\phi`) instead of a displacement.

4. **Finally, we measure the mean photon number** :math:`\langle\hat{n}\rangle =
   \langle\hat{a}^\dagger \hat{a}\rangle`. This quantity, which tells us the average amount of photons in the final state, is proportional to the energy of the photonic system.

The aim of this tutorial is to optimize the circuit parameters :math:`(\alpha, \phi)`
such that the mean photon number is equal to one. The rotation gate is actually a
*passive transformation*, meaning that it does not change the energy of the system.
The displacement gate is an *active transformation*, which modifies the energy of the
photonic system.
"""


###############################################################################
# Constructing the QNode
# ----------------------
#
# As before, we import PennyLane, as well as the wrapped version of NumPy provided
# by PennyLane:

import pennylane as qml
from pennylane import numpy as np

###############################################################################
# Next, we instantiate a device which will be used to evaluate the circuit.
# Because our circuit contains only Gaussian operations, we can make use of the
# built-in ``default.gaussian`` device.

dev_gaussian = qml.device("default.gaussian", wires=1)

###############################################################################
# After initializing the device, we can construct our quantum node. As before, we use the
# :mod:`qnode decorator <pennylane.decorator>` to convert our quantum function
# (encoded by the circuit above) into a quantum node running on the ``default.gaussian``
# device.


@qml.qnode(dev_gaussian)
def mean_photon_gaussian(mag_alpha, phase_alpha, phi):
    qml.Displacement(mag_alpha, phase_alpha, wires=0)
    qml.Rotation(phi, wires=0)
    return qml.expval(qml.NumberOperator(0))


###############################################################################
# Notice that we have broken up the complex number :math:`\alpha` into two real
# numbers ``mag_alpha`` and ``phase_alpha``, which form a polar representation of
# :math:`\alpha`. This is so that the notion of a gradient is clear and well-defined.


###############################################################################
# Optimization
# ------------
#
# As in the :ref:`qubit rotation <qubit_rotation>` tutorial, let's now use one
# of the built-in PennyLane optimizers in order to optimize the quantum circuit
# towards the desired output. We want the mean photon number to be exactly one,
# so we will use a squared-difference cost function:


def cost(params):
    return (mean_photon_gaussian(params[0], params[1], params[2]) - 1.0) ** 2


###############################################################################
# At the beginning of the optimization, we choose arbitrary small initial parameters:

init_params = [0.015, 0.02, 0.005]
print(cost(init_params))

###############################################################################
# When the gate parameters are near to zero, the gates are close to the
# identity transformation, which leaves the initial state largely unchanged.
# Since the initial state contains no photons, the mean photon number of the
# circuit output is approximately zero, and the cost is close to one.

###############################################################################
# .. note::
#
#     We avoided initial parameters which are exactly zero because that
#     corresponds to a critical point with zero gradient.
#
# Now, let's use the :class:`~.GradientDescentOptimizer`, and update the circuit
# parameters over 100 optimization steps.

# initialise the optimizer
opt = qml.GradientDescentOptimizer(stepsize=0.1)

# set the number of steps
steps = 20
# set the initial parameter values
params = init_params

for i in range(steps):
    # update the circuit parameters
    params = opt.step(cost, params)

    print("Cost after step {:5d}: {:8f}".format(i + 1, cost(params)))

print("Optimized mag_alpha:{:8f}".format(params[0]))
print("Optimized phase_alpha:{:8f}".format(params[1]))
print("Optimized phi:{:8f}".format(params[2]))

###############################################################################
# The optimization converges after about 20 steps to a cost function value
# of zero.
#
# We observe that the two angular parameters ``phase_alpha`` and ``phi``
# do not change during the optimization. Only the magnitude of the complex
# displacement :math:`|\alpha|` affects the mean photon number of the circuit.
#
# Continue on to the next tutorial, :ref:`plugins_hybrid`, to learn how to
# utilize the extensive plugin ecosystem of PennyLane,
# build continuous-variable (CV) quantum nodes, and to see an example of a
# hybrid qubit-CV-classical computation using PennyLane.



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultGaussian` device.
"""
# pylint: disable=protected-access,cell-var-from-loop
from pennylane import numpy as np
from scipy.linalg import block_diag

import pennylane as qml
from pennylane.qnode import QuantumFunctionError
from pennylane.plugins import DefaultQubit

import pytest




def test_pass_positional_wires_to_observable(monkeypatch, capfd):
    """Tests whether the ability to pass wires as positional argument is retained"""
    dev = qml.device("default.qubit", wires=1)

    @qml.qnode(dev)
    def circuit():
        return qml.expval(qml.Identity(0))

    with monkeypatch.context() as m:
        m.setattr(DefaultQubit, "pre_measure", lambda self: print(self.obs_queue))
        circuit()

    out, err = capfd.readouterr()
    assert "pennylane.ops.Identity object" in out
